"""
pyart.io._sigmetfile
====================

A class and supporting functions for reading Sigmet (raw format) files.

.. autosummary::
    :toctree: generated/

    SigmetFile
    convert_sigmet_data
    bin2_to_angle
    bin4_to_angle
    _data_types_from_mask
    _is_bit_set
    _parse_ray_headers
    _unpack_structure
    _unpack_key
    _unpack_ingest_data_headers
    _unpack_ingest_data_header
    _unpack_raw_prod_bhdr
    _unpack_product_hdr
    _unpack_ingest_header

"""
import struct
import datetime
import warnings

import numpy as np
cimport numpy as np
cimport cython

RECORD_SIZE = 6144      # Raw product file blocked into 6144 byte records


cdef class SigmetFile:
    """
    A class for accessing data from Sigmet (IRIS) product files.

    Parameters
    ----------
    filename : str
        Filename or file-like object.

    Attributes
    ----------
    debug : bool
        Set True to print out debugging information, False otherwise.
    product_hdr : dict
        Product_hdr structure.
    ingest_header : dict
        Ingest_header structure.
    ingest_data_headers : list of dict
        Ingest_data_header structures for each data type.  Indexed by the
        data type name (str).  None when data has not yet been read.
    data_types : list
        List of data types (int) in the file.
    data_type_names : list
        List of data type names (stR) in the file.
    ndata_types : int
        Number of data types in the file.
    _fh : file
        Open file being read.
    _raw_product_bhdrs : list
        List of raw_product_bhdr structure dictionaries seperated by sweep.
        None when data has not yet been read.

    """
    cdef public debug, product_hdr, ingest_header, ingest_data_headers, \
        data_types, data_type_names, ndata_types,
    cdef public _fh, _raw_product_bhdrs

    cdef np.ndarray _rbuf
    cdef np.int16_t * _rbuf_p   # hack for fast indexing of _rbuf
    cdef public int _rbuf_pos, _record_number

    def __init__(self, filename, debug=False):
        """ initalize the object. """

        self.debug = debug

        # open the file
        if hasattr(filename, 'read'):
            fh = filename
        else:
            fh = open(filename, 'rb')

        # read the headers from the first 2 records.
        self.product_hdr = _unpack_product_hdr(fh.read(RECORD_SIZE))
        self.ingest_header = _unpack_ingest_header(fh.read(RECORD_SIZE))

        # determine data types contained in the file
        self.data_types = self._determine_data_types()
        self.ndata_types = len(self.data_types)
        self.data_type_names = [SIGMET_DATA_TYPES[i] for i in self.data_types]

        # set attributes
        self.ingest_data_headers = None
        self._fh = fh
        self._record_number = 2
        self._raw_product_bhdrs = []

    def _determine_data_types(self):
        """ Determine the available data types in the file. """
        # determine the available fields
        task_config = self.ingest_header['task_configuration']
        task_dsp_info = task_config['task_dsp_info']
        word0 = task_dsp_info['current_data_type_mask']['mask_word_0']
        word1 = task_dsp_info['current_data_type_mask']['mask_word_1']
        word2 = task_dsp_info['current_data_type_mask']['mask_word_2']
        word3 = task_dsp_info['current_data_type_mask']['mask_word_3']
        return _data_types_from_mask(word0, word1, word2, word3)

    def close(self):
        """ Close the file. """
        self._fh.close()

    def read_data(self, full_xhdr=False):
        """
        Read all data from the file.

        Parameters
        ----------
        full_xhdr : bool
            True to return the full extended headers if they exist padded with
            ones.  False will return a length 1 extended header converted to
            int32.  This is useful when the file contains a customer specified
            extended header (for example aircraft radar).

        Returns
        -------
        data : dict of ndarrays
            Data arrays of shape=(nsweeps, nrays, nbins) for each data type.
            Indexed by data type name (str).
        metadata : dict of dicts
            Arrays of 'azimuth_0', 'azimuth_1', 'elevation_0', 'elevation_1',
            'nbins', and 'time' for each data type.  Indexed by data type name
            (str).  Rays which were not collected are marked with a value of
            -1 in the 'nbins' array.

        """

        # determine size of data
        nsweeps = self.ingest_header['task_configuration'][
            'task_scan_info']['number_sweeps']
        nbins = self.product_hdr['product_end']['number_bins']
        nrays = self.ingest_header['ingest_configuration'][
            'number_rays_sweep']

        # create empty outputs
        shape = (nsweeps, nrays, nbins)
        data = dict([(name, np.ma.empty(shape, dtype='float32'))
                    for name in self.data_type_names])
        if 'XHDR' in self.data_type_names:
            if full_xhdr:
                data['XHDR'] = np.ones(shape, dtype='int16')
            else:
                data['XHDR'] = np.ones((nsweeps, nrays, 1), dtype='int32')

        metadata = {}
        for name in self.data_type_names:
            header_dic = {
                'azimuth_0': np.empty((nsweeps, nrays), dtype='float32'),
                'elevation_0': np.empty((nsweeps, nrays), dtype='float32'),
                'azimuth_1': np.empty((nsweeps, nrays), dtype='float32'),
                'elevation_1': np.empty((nsweeps, nrays), dtype='float32'),
                'nbins': np.empty((nsweeps, nrays), dtype='int16'),
                'time': np.empty((nsweeps, nrays), dtype='uint16'),
                'prf_flag': np.empty((nsweeps, nrays), dtype='int16')}
            metadata[name] = header_dic

        self.ingest_data_headers = dict([(name, []) for name in
                                         self.data_type_names])

        self._raw_product_bhdrs = []

        # read in data sweep by sweep
        for i in xrange(nsweeps):
            ingest_data_hdrs, sweep_data, sweep_metadata = self._get_sweep(
                full_xhdr=full_xhdr)

            # check for a truncated file, return sweep(s) read up until error
            if ingest_data_hdrs is None:

                mess = ('File truncated or corrupt, %i of %i sweeps read' %
                        (i, nsweeps))
                warnings.warn(mess)

                for name in self.data_type_names:
                    data[name] = data[name][:i]
                    for k in metadata[name]:
                        metadata[name][k] = metadata[name][k][:i]
                return data, metadata

            for j, name in enumerate(self.data_type_names):
                temp = sweep_metadata[j]
                (az0, el0, az1, el1, ray_nbins, ray_time, prf_flag) = temp

                data[name][i] = sweep_data[j]
                metadata[name]['azimuth_0'][i] = az0
                metadata[name]['azimuth_1'][i] = az1
                metadata[name]['elevation_0'][i] = el0
                metadata[name]['elevation_1'][i] = el1
                metadata[name]['nbins'][i] = ray_nbins
                metadata[name]['time'][i] = ray_time
                metadata[name]['prf_flag'][i] = prf_flag
                self.ingest_data_headers[name].append(ingest_data_hdrs[j])

        # scale 1-byte velocity by the Nyquist (section 4.3.29)
        # this conversion is kept in this method so that the
        # product_hdr does not need to be accessed at lower abstraction
        # layers.
        if 'VEL' in self.data_type_names:
            wavelength_cm = self.product_hdr['product_end']['wavelength']
            prt_value = 1. / self.product_hdr['product_end']['prf']
            task_config = self.ingest_header['task_configuration']
            multi_prf_flag = task_config['task_dsp_info']['multi_prf_flag']
            if multi_prf_flag > 3 or multi_prf_flag < 0:
                multiplier = 1  # multiplier not defined in IRIS manual
            else:
                multiplier = [1, 2, 3, 4][multi_prf_flag]
            nyquist = wavelength_cm / (10000.0 * 4.0 * prt_value) * multiplier
            data['VEL'] *= nyquist
        # scale 1-byte width by the Nyquist
        if 'WIDTH' in self.data_type_names:
            # The IRIS Programmer's Manual indicates 1-byte width format data
            # should be scaled by the unambiguous velocity, twice the nyquist,
            # (section 4.3.35) but both RSL and RadX scale this data by the
            # nyquist.  Therefore to agree with these two packages the width
            # is scaled by the nyquist.
            wavelength_cm = self.product_hdr['product_end']['wavelength']
            prt_value = 1. / self.product_hdr['product_end']['prf']
            nyquist = wavelength_cm / (10000.0 * 4.0 * prt_value)
            data['WIDTH'] *= nyquist

        return data, metadata

    def _get_sweep(self, full_xhdr=False, raw_data=False):
        """
        Get the data and metadata from the next sweep.

        If the file ends early None is returned for all values.

        Parameters
        ----------
        full_xhdr : bool
            True to return the full extended headers if they exist padded with
            ones.  False will return a length 1 extended header converted to
            int32.  This is useful when the file contains a customer specified
            extended header (for example aircraft radar).
        raw_data : bool, optional
            True to return the raw_data for the given sweep, False to
            convert the data to floating point representation.

        Returns
        -------
        ingest_data_headers : list of dict
            List of ingest_data_header structures for each data type.
        sweep_data : list of arrays
            Sweep data for each data types in the order they appear in the
            file.
        sweep_metadata : list of tuples
            Sweep metadata for each data type in the same order as sweep_data.

        """

        # get the next record
        lead_record = self._fh.read(RECORD_SIZE)
        self._record_number += 1

        # check if the file ended early, if so return Nones
        if len(lead_record) != RECORD_SIZE:
            return None, None, None

        # unpack structures
        raw_prod_bhdr = _unpack_raw_prod_bhdr(lead_record)
        self._raw_product_bhdrs.append([raw_prod_bhdr])
        ingest_data_headers = _unpack_ingest_data_headers(
            lead_record, self.ndata_types)
        if ingest_data_headers is None:
            return None, None, None

        # determine size of data
        nray_data_types = [d['number_rays_file_expected']
                           for d in ingest_data_headers]
        nrays = sum(nray_data_types)    # total rays
        nbins = self.product_hdr['product_end']['number_bins']

        # prepare to read rays
        self._rbuf = np.frombuffer(lead_record, dtype='int16')
        self._rbuf_p = <np.int16_t*>self._rbuf.data
        self._rbuf_pos = int((12 + 76 * self.ndata_types) / 2) - 1
        # set data initially to ones so that missing data can be better
        # seen when debugging
        raw_sweep_data = np.ones((nrays, nbins + 6), dtype='int16')

        # get the raw data ray-by-ray
        for ray_i in xrange(nrays):
            if self.debug:
                print("Reading ray: %i of %i" % (ray_i, nrays))
                print("self._rbuf_pos is", self._rbuf_pos)
            if self._get_ray(nbins, raw_sweep_data[ray_i]):
                return None, None, None

        # return raw data if requested
        if raw_data:
            return ingest_data_headers, raw_sweep_data

        # convert the data and parse the metadata
        sweep_data = []
        sweep_metadata = []
        for i, data_type in enumerate(self.data_types):
            if data_type == 0 and full_xhdr:
                sweep_data.append(raw_sweep_data[i::self.ndata_types, 6:])
            else:
                sweep_data.append(convert_sigmet_data(
                    data_type, raw_sweep_data[i::self.ndata_types, 6:],
                    raw_sweep_data[i::self.ndata_types, 4]))
            sweep_metadata.append(_parse_ray_headers(
                raw_sweep_data[i::self.ndata_types, :6]))
        return ingest_data_headers, sweep_data, sweep_metadata

    @cython.wraparound(False)
    cdef int _get_ray(self, int nbins, np.ndarray[np.int16_t, ndim=1] out):
        """
        Get the next ray, loading new records as needed.

        Parameters
        ----------
        nbins : int
            Number of bins in the ray.
        out : ndarray
            Array to load ray data into.

        Returns
        -------
        status : int
            0 on success, -1 if failed.

        """
        cdef short compression_code
        cdef int words, remain, out_pos, first_end, i

        if self._incr_rbuf_pos():
            return -1   # failed read
        compression_code = self._rbuf_p[self._rbuf_pos]
        out_pos = 0

        if compression_code == 1:
            # mark ray as missing by setting numbers of bins to -1
            out[4] = -1
            return 0

        while compression_code != 1:

            if self._incr_rbuf_pos():
                return -1   # failed read
            if compression_code < 0:
                words = compression_code + 32768    # last 7 bits give size
                if self._rbuf_pos + words <= 3072:
                    # all compressed data is in the current record
                    for i in range(words):
                        out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]
                    if self._incr_rbuf_pos(words):
                        return -1   # failed read
                    out_pos += words
                else:
                    # data is split between current and next record
                    # store data from current record
                    remain = words - (3072 - self._rbuf_pos)
                    first_end = out_pos + words - remain
                    for i in range(first_end - out_pos):
                        out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]

                    # read data from next record and store
                    self._load_record()
                    for i in range(out_pos + words - first_end):
                        out[first_end + i] = self._rbuf_p[self._rbuf_pos + i]

                    if self._incr_rbuf_pos(remain):
                        return -1   # failed read
                    out_pos += words
            else:
                # add zeros to out
                if compression_code + out_pos > nbins + 6:
                    return -1   # file is corrupt
                for i in range(compression_code):
                    out[out_pos + i] = 0
                out_pos += compression_code
            compression_code = self._rbuf_p[self._rbuf_pos]

        return 0

    cdef int _incr_rbuf_pos(self, int incr=1):
        """
        Increment the record buffer position, load a new record if needed.
        """
        self._rbuf_pos += incr
        if self._rbuf_pos >= 3072:
            if self._load_record():
                return -1   # failed read
        return 0

    cdef int _load_record(self):
        """ Load the next record. returns -1 on fail, 0 if success. """
        record = self._fh.read(RECORD_SIZE)
        if len(record) != RECORD_SIZE:
            return -1   # failed read
        self._record_number += 1
        if self.debug:
            print("Finished loading record:", self._record_number)
        self._raw_product_bhdrs[-1].append(_unpack_raw_prod_bhdr(record))
        self._rbuf = np.frombuffer(record, dtype='int16')
        self._rbuf_pos = 6
        self._rbuf_p = <np.int16_t*>self._rbuf.data
        return 0

# functions used by the SigmetFile class


def _data_types_from_mask(word0, word1, word2, word3):
    """
    Return a list of the data types from the words in the data_type mask.
    """
    data_types = [i for i in range(32) if _is_bit_set(word0, i)]
    data_types += [i+32 for i in range(32) if _is_bit_set(word1, i)]
    data_types += [i+64 for i in range(32) if _is_bit_set(word2, i)]
    data_types += [i+96 for i in range(32) if _is_bit_set(word3, i)]
    return data_types


def _is_bit_set(number, bit):
    """ Return True if bit is set in number. """
    return number >> bit & 1 == 1


def _parse_ray_headers(ray_headers):
    """
    Parse the metadata from Sigmet ray headers.

    Parameters
    ----------
    ray_headers : array, shape=(..., 6)
        Ray headers to parse.

    Returns
    -------
    az0 : array
        Azimuth angles (in degrees) at beginning of the rays.
    el0 : array
        Elevation angles at the beginning of the rays.
    az1 : array
        Azimuth angles at the end of the rays.
    el1 : array
        Elevation angles at the end of the rays.
    nbins : array
        Number of bins in the rays.
    time : array
        Seconds since the start of the sweep for the rays.
    prf_flag : array
        Numerical indication of what PRF was used, 0 for high, 1 for low.
        Not applicable if dual-PRF is not used during collection.

    """
    az0 = bin2_to_angle(ray_headers.view('uint16')[..., 0])
    el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])
    az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
    el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
    nbins = ray_headers.view('int16')[..., 4]
    time = ray_headers.view('uint16')[..., 5]
    prf_flag = np.mod(ray_headers.view('int16')[..., 0], 2)
    return (az0, el0, az1, el1, nbins, time, prf_flag)


###################
# format converts #
###################

# Data type constants, table 13, section 4.8
SIGMET_DATA_TYPES = {
    0: 'XHDR',
    1: 'DBT',
    2: 'DBZ',
    3: 'VEL',
    4: 'WIDTH',
    5: 'ZDR',
    6: 'UNKNOWN_6',     # Not known
    7: 'DBZC',
    8: 'DBT2',
    9: 'DBZ2',
    10: 'VEL2',
    11: 'WIDTH2',
    12: 'ZDR2',
    13: 'RAINRATE2',
    14: 'KDP',
    15: 'KDP2',
    16: 'PHIDP',
    17: 'VELC',
    18: 'SQI',
    19: 'RHOHV',
    20: 'RHOHV2',
    21: 'DBZC2',
    22: 'VELC2',
    23: 'SQI2',
    24: 'PHIDP2',
    25: 'LDRH',
    26: 'LDRH2',
    27: 'LDRV',
    28: 'LDRV2',
    29: 'UNKNOWN_29',   # Not known
    30: 'UNKNOWN_30',   # Not known
    31: 'UNKNOWN_31',   # Not known
    32: 'HEIGHT',
    33: 'VIL2',
    34: 'RAW',
    35: 'SHEAR',
    36: 'DIVERGE2',
    37: 'FLIQUID2',
    38: 'USER',
    39: 'OTHER',
    40: 'DEFORM2',
    41: 'VVEL2',
    42: 'HVEL2',
    43: 'HDIR2',
    44: 'AXDIL2',
    45: 'TIME2',
    46: 'RHOH',
    47: 'RHOH2',
    48: 'RHOV',
    49: 'RHOV2',
    50: 'PHIH',
    51: 'PHIH2',
    52: 'PHIV',
    53: 'PHIV2',
    54: 'USER2',
    55: 'HCLASS',
    56: 'HCLASS2',
    57: 'ZDRC',
    58: 'ZDRC2',
    59: 'TEMPERATURE16',
    60: 'VIR16',
    61: 'DBTV8',
    62: 'DBTV16',
    63: 'DBZV8',
    64: 'DBZV16',
    65: 'SNR8',
    66: 'SNR16',
    67: 'ALBEDO8',
    68: 'ALBEDO16',
    69: 'VILD16',
    70: 'TURB16',
    71: 'DBTE8',
    72: 'DBTE16',       # Total Power Enhanced
    73: 'DBZE8',
    74: 'DBZE16',       # Clutter Corrected Reflectivity Enhanced
    # Uknown fields, do not know internal names, some may be user defined.
    75: 'UNKNOWN_75',
    76: 'UNKNOWN_76',
    77: 'UNKNOWN_77',
    78: 'UNKNOWN_78',
    79: 'UNKNOWN_79',
    80: 'UNKNOWN_80',
    81: 'UNKNOWN_81',
    82: 'UNKNOWN_82',
    83: 'UNKNOWN_83',
    84: 'UNKNOWN_84',
    85: 'UNKNOWN_85',
    86: 'UNKNOWN_86',
    87: 'UNKNOWN_87',
    88: 'UNKNOWN_88',
    89: 'UNKNOWN_89',
    90: 'UNKNOWN_90',
    91: 'UNKNOWN_91',
    92: 'UNKNOWN_92',
    93: 'UNKNOWN_93',
    94: 'UNKNOWN_94',
    95: 'UNKNOWN_95',
    96: 'UNKNOWN_96',
    97: 'UNKNOWN_97',
    98: 'UNKNOWN_98',
    99: 'UNKNOWN_99',
    100: 'UNKNOWN_100',
    101: 'UNKNOWN_101',
    102: 'UNKNOWN_102',
    103: 'UNKNOWN_103',
    104: 'UNKNOWN_104',
    105: 'UNKNOWN_105',
    106: 'UNKNOWN_106',
    107: 'UNKNOWN_107',
    108: 'UNKNOWN_108',
    109: 'UNKNOWN_109',
    110: 'UNKNOWN_110',
    111: 'UNKNOWN_110',
    112: 'UNKNOWN_112',
    113: 'UNKNOWN_113',
    114: 'UNKNOWN_114',
    115: 'UNKNOWN_115',
    116: 'UNKNOWN_116',
    117: 'UNKNOWN_117',
    118: 'UNKNOWN_118',
    119: 'UNKNOWN_119',
    120: 'UNKNOWN_120',
    121: 'UNKNOWN_121',
    122: 'UNKNOWN_122',
    123: 'UNKNOWN_123',
    124: 'UNKNOWN_124',
    125: 'UNKNOWN_125',
    126: 'UNKNOWN_126',
    127: 'UNKNOWN_127',
}   # there may be more field, add as needed


# This function takes a majority of the time when reading data from a sigmet
# file. Rewriting the convertions/masking in Cython does not seem to improved
# performance likely since most of the routines are already vectorized.
def convert_sigmet_data(data_type, data, nbins):
    """ Convert sigmet data. """
    out = np.empty_like(data, dtype='float32')
    mask = np.zeros_like(data, dtype=np.bool8)

    data_type_name = SIGMET_DATA_TYPES[data_type]

    like_dbt2 = [
        'DBT2',     # 2-byte Reflectivity Format, section 4.3.4
        'DBZ2',     # " "
        'KDP2',     # 2-byte KDP Format, section 4.3.13
        'LDRH2',    # 2-byte LDR Format, section 4.3.15
        'LDRV2',    # " "
        'VEL2',     # 2-byte Velocity Format, section 4.3.30
        'VELC2',    # 2-byte Unfolded Velocity Format, section 4.3.32
        'ZDR2',     # 2-byte ZDR Format, section 4.3.38
        'DBZC2',    # Corrected reflectivity, XXX not certain of format
        'ZDRC2',    # Corrected differential reflectivity, XXX not certain
        'DBTV16',   # Total V power, 2-byte
        'DBZV16',   # Clutter corrected V reflectivity, 2-byte
        'SNR16',    # Signal to noise ratio, 2-byte
        'DBTE16',   # Total Power Enhanced, 2-byte
        'DBZE16',   # Clutter corrected reflectivity enhanced, 2-byte
    ]

    like_sqi = [
        'RHOH',     # 1-byte Rho Format, section 4.3.21
        'RHOV',     # " "
        'RHOHV',    # 1-byte RhoHV Format, section 4.3.23
        'SQI',      # 1-byte Signal Quality Index Format, section 4.3.26
    ]

    like_sqi2 = [
        'RHOV2',    # 2-byte Rho Format, section 4.3.22
        'RHOH2',    # " "
        'RHOHV2',   # 2-byte RhoHV Format, section 4.3.24
        'SQI2',     # 2-byte Signal Quality Index Format, section 4.3.27
    ]

    like_dbt = [
        'DBT',      # 1-bytes Reflectivity Format, section 4.3.3
        'DBZ',      # " "
        'DBTV8',    # Total V power, 1-byte
        'DBZV8',    # Clutter corrected V reflectivity, 1-byte
        'SNR8',     # Signal to noise ratio, 1-byte
        'DBTE8',    # Total power enhanced, 1-byte
        'DBZE8',    # Clutter corrected reflectivity enhanced, 1-byte
    ]

    if data_type_name in like_dbt2:
        # value = (N - 32768) / 100.
        # 0 : no data available (mask)
        # 65535 Reserved for area not scanned in product file (nothing)
        out[:] = (data.view('uint16') - 32768.) / 100.
        mask[data.view('uint16') == 0] = True

    elif data_type_name in like_sqi2:
        # value = (N - 1) / 65533
        # 0 : no data available (mask)
        # 65535 Area not scanned
        out[:] = (data.view('uint16') - 1.) / 65533.
        mask[data.view('uint16') == 0] = True

    elif data_type_name == 'WIDTH2':
        # DB_WIDTH2, 11, Width (2 byte)
        # 2-byte Width Format, section 4.3.36
        out[:] = data.view('uint16') / 100.
        mask[data.view('uint16') == 0] = True

    elif data_type_name == 'PHIDP2':
        # DB_PHIDP2, 24, PhiDP (Differential Phase) (2 byte)
        # 2-byte PhiDP format, section 4.3.19
        out[:] = 360. * (data.view('uint16') - 1.) / 65534.
        mask[data.view('uint16') == 0] = True

    elif data_type_name == 'HCLASS2':
        # DB_HCLASS2, 56, Hydrometeor class (2 byte)
        # 2-byte HydroClass Format, section 4.3.9
        out[:] = data.view('uint16')

    elif data_type_name == 'XHDR':
        # Extended Headers, 0
        # extended_header_v0, _v1, _v2, section 4.2.8-4.2.10
        # Here we return an array with the times in milliseconds.
        return data[..., :2].copy().view('i4')

    # one byte data types
    elif data_type_name[-1] != '2':
        # make a view of left half of the data as uint8,
        # this is the actual ray data collected, the right half is blank.
        nrays, nbin = data.shape
        ndata = data.view('(2,) uint8').reshape(nrays, -1)[:, :nbin]

        if data_type_name in like_dbt:
            # DB_DBT, 1, Total Power (1 byte)
            # 1-byte Reflectivity Format, section 4.3.3
            out[:] = (ndata - 64.) / 2.
            mask[ndata == 0] = True

        elif data_type_name in like_sqi:
            # value = sqrt((N - 1) / 253)
            # 0 : no data available (mask)
            # 255 Area not scanned
            out[:] = np.sqrt((ndata - 1.) / 253.)
            mask[ndata == 0] = True
            mask[ndata == 255] = True

        elif data_type_name == 'VEL':
            # VEL, 3, Velocity (1 byte)
            # 1-byte Velocity Format, section 4.3.29
            # Note that this data should be multiplied by Nyquist,
            # this is done in the get_data method of the SigmetFile class.
            out[:] = (ndata - 128.) / 127.
            mask[ndata == 0] = True

        elif data_type_name == 'WIDTH':
            # WIDTH, 4, Width (1 byte)
            # 1-byte Width format, section 4.3.25
            # Note that this data should be multiplied by the unambiguous
            # velocity
            out[:] = ndata / 256.
            mask[ndata == 0] = True

        elif data_type_name == 'ZDR':
            # ZDR, 5, Differential reflectivity (1 byte)
            # 1-byte ZDR format, section 4.3.37
            out[:] = (ndata - 128.) / 16.
            mask[ndata == 0] = True

        elif data_type_name == 'KDP':
            # KDP, 14, KDP (Differential phase) (1 byte)
            # 1-byte KDP format, section 4.3.12
            # Note that this data should be divided by the wavelength in cm
            # as is the units are deg * cm / km

            # above 128 use positive value equation
            exp = np.power(600., (ndata[ndata > 128] - 129.) / 126.)
            out[ndata > 128] = 0.25 * exp
            # below 128, use negative value equation
            exp = np.power(600., (127. - ndata[ndata < 128]) / 126.)
            out[ndata < 128] = -0.25 * exp
            # equal to 128, zero
            out[ndata == 128] = 0

            mask[ndata == 0] = True
            mask[ndata == 255] = True

        elif data_type_name == 'PHIDP':
            # PHIDP, 16, PhiDP(Differential phase) (1 byte)
            # 1-byte PhiDP format, section 4.3.18
            out[:] = 180. * ((ndata - 1.) / 254.)
            mask[ndata == 0] = True
            mask[ndata == 255] = True

        elif data_type_name == "HCLASS":
            # HCLASS, 55, Hydrometeor class (1 byte)
            # 1-byte HydroClass format, section 4.3.8
            out[:] = ndata[:]
            mask[ndata == 0] = True     # No data available
            mask[ndata == 255] = True   # Area not scanned

        else:
            # TODO implement conversions for addition 1-byte formats
            warnings.warn('Unknown type: %s, returning raw data' % data_type)
            out[:] = np.ma.masked_array(data)
            return out
    else:
        # TODO implement conversions for additional formats.
        warnings.warn('Unknown type: %s, returning raw data' % data_type)
        out[:] = data
        return np.ma.masked_array(out)

    # mask any gates which are beyond the number of gates in that ray.
    _mask_gates_not_collected(mask.view(np.uint8), nbins)

    return np.ma.masked_array(out, mask=mask, fill_value=-9999.0,
                              shrink=False)


@cython.boundscheck(False)
cdef _mask_gates_not_collected(
        np.ndarray[np.uint8_t, ndim=2] mask,
        np.ndarray[np.int16_t, ndim=1] nbins):
    """ Add gates not collected (beyond nbin) to the mask. """
    cdef int i, j, nrays, nbin, full_nbins
    nrays = mask.shape[0]
    full_nbins = mask.shape[1]
    for i in range(nrays):
        nbin = nbins[i]
        for j in range(nbin, full_nbins):
            mask[i, j] = 1
    return


def bin2_to_angle(bin2):
    """ Return an angle from Sigmet bin2 encoded value (or array). """
    return 360. * bin2 / 65536


def bin4_to_angle(bin4):
    """ Return an angle from Sigmet bin4 encoded value (or array). """
    return 360. * bin4 / 4294967296


#####################
# get/put functions #
#####################


def _unpack_structure(string, structure):
    """ Unpack a structure """
    fmt = ''.join([i[1] for i in structure])
    l = struct.unpack(fmt, string)
    return dict(zip([i[0] for i in structure], l))


def _unpack_key(dic, key, structure):
    """ Unpack a key. """
    dic[key] = _unpack_structure(dic[key], structure)


def _unpack_ingest_data_headers(record, ndata_types):
    """
    Unpack one or more ingest_data_header from a record.

    Returns a list of dictionaries or None when an error occurs.

    """
    idh = [_unpack_ingest_data_header(record, i) for i in range(ndata_types)]
    if None in idh:
        return None
    else:
        return idh


def _unpack_ingest_data_header(record, number):
    """
    Unpack a single ingest_data_header from record.  Return None on error.
    """
    offset = 12 + 76 * number
    string = record[offset:offset + 76]
    idh = _unpack_structure(string, INGEST_DATA_HEADER)
    _unpack_key(idh, 'structure_header', STRUCTURE_HEADER)
    _unpack_key(idh, 'sweep_start_time', YMDS_TIME)
    if idh['structure_header']['structure_identifier'] != 24:
        return None
    return idh


def _unpack_raw_prod_bhdr(record):
    """ Return a dict with the unpacked raw_prod_bhdr from a record. """
    return _unpack_structure(record[:12], RAW_PROD_BHDR)


def _unpack_product_hdr(record):
    """
    Return a dict with the unpacked product_hdr from the first record.
    """

    # unpack the product_hdr structure from the first record
    product_hdr = _unpack_structure(record[:640], PRODUCT_HDR)

    # product_hdr substructure
    _unpack_key(product_hdr, 'structure_header', STRUCTURE_HEADER)
    _unpack_key(product_hdr, 'product_configuration',
                PRODUCT_CONFIGURATION)
    _unpack_key(product_hdr, 'product_end', PRODUCT_END)

    # product_config substructure
    product_config = product_hdr['product_configuration']
    _unpack_key(product_config, 'structure_header', STRUCTURE_HEADER)
    _unpack_key(product_config, 'generation_time', YMDS_TIME)
    _unpack_key(product_config, 'sweep_ingest_time', YMDS_TIME)
    _unpack_key(product_config, 'file_ingest_time', YMDS_TIME)
    _unpack_key(product_config, 'color_scale_def', COLOR_SCALE_DEF)

    # product_end substructure
    product_end = product_hdr['product_end']
    _unpack_key(product_end, 'ingest_time', YMDS_TIME)

    return product_hdr


def _unpack_ingest_header(record):
    """
    Return a dict with the unpacked ingest_header from the second record.
    """

    # unpack the ingest_header structure from the second_record
    ingest_header = _unpack_structure(record[:4884], INGEST_HEADER)

    # ingest_header substructure
    _unpack_key(ingest_header, 'structure_header', STRUCTURE_HEADER)
    _unpack_key(ingest_header, 'ingest_configuration',
                INGEST_CONFIGURATION)
    _unpack_key(ingest_header, 'task_configuration', TASK_CONFIGURATION)

    # ingest_configuration substructure
    ingest_configuration = ingest_header['ingest_configuration']
    _unpack_key(ingest_configuration, 'volume_scan_start_time', YMDS_TIME)

    # task_configuration substructure
    task_configuration = ingest_header['task_configuration']
    _unpack_key(task_configuration, 'structure_header', STRUCTURE_HEADER)
    _unpack_key(task_configuration, 'task_sched_info', TASK_SCHED_INFO)
    _unpack_key(task_configuration, 'task_dsp_info', TASK_DSP_INFO)
    _unpack_key(task_configuration, 'task_calib_info', TASK_CALIB_INFO)
    _unpack_key(task_configuration, 'task_range_info', TASK_RANGE_INFO)
    _unpack_key(task_configuration, 'task_scan_info', TASK_SCAN_INFO)
    _unpack_key(task_configuration, 'task_misc_info', TASK_MISC_INFO)
    _unpack_key(task_configuration, 'task_end_info', TASK_END_INFO)

    # task_dsp_info substructure
    task_dsp_info = task_configuration['task_dsp_info']
    _unpack_key(task_dsp_info, 'current_data_type_mask', DSP_DATA_MASK)
    _unpack_key(task_dsp_info, 'original_data_type_mask', DSP_DATA_MASK)
    _unpack_key(task_dsp_info, 'task_dsp_mode', TASK_DSP_MODE_BATCH)

    # task_scan_info substructure
    # TODO unpack task_scan_type_scan_info based on scan type
    # task_scan_info = task_configuration['task_scan_info']
    #    scan_type_struct =
    # _unpack_key(task_scan_info, 'task_scan_type_scan_info',
    #            scan_type_struct)

    # task_end_info substructure
    task_end_info = task_configuration['task_end_info']
    _unpack_key(task_end_info, 'task_data_time', YMDS_TIME)

    return ingest_header


##############
# structures #
##############

# scalar defitions, section 4.1, table 7, and corresponding
SINT1 = 'b'
UINT1 = 'B'
SINT2 = 'h'
UINT2 = 'H'
SINT4 = 'i'
UINT4 = 'I'
FLT4 = 'f'
FLT8 = 'd'
BIN1 = 'B'
BIN2 = 'H'      # these values need to be decoded with _bin2_to_angle
BIN4 = 'I'      # these values need to be decoded with _bin4_to_angle
MESSAGE = 'I'
UINT16_T = 'H'

# structures are taken from Vaisala PROGRAMMER'S MANUAL IRIS
# M21131EN-B

# Chapter 4 deals with data formats
# section 4.5.4 deals with the RAW Product format

# 640 bytes: product_hdr (section 4.2.25, page 47)
PRODUCT_HDR = (
    ('structure_header', '12s'),        # 12 bytes
    ('product_configuration', '320s'),  # 320 bytes
    ('product_end', '308s'),            # 308 bytes
)

# 12 bytes : structure_header (section 4.2.47)
STRUCTURE_HEADER = (
    ('structure_identifier', SINT2),
    ('format_version', SINT2),
    ('bytes_in_structure', SINT4),
    ('reserved', SINT2),
    ('flag', SINT2),
)

# 320 bytes: product_configuration (section 4.2.23, page 43) 320 bytes
PRODUCT_CONFIGURATION = (
    ('structure_header', '12s'),    # 12 bytes: structure_header
    ('product_type_code', UINT2),
    ('scheduling_code', UINT2),
    ('seconds_between_runs', SINT4),
    ('generation_time', '12s'),     # 12 bytes: ymds_time
    ('sweep_ingest_time', '12s'),   # 12 bytes: ymds_time
    ('file_ingest_time', '12s'),    # 12 bytes: ymds_time
    ('spare_0', '6s'),              # 6 bytes
    ('product_name', '12s'),
    ('task_name', '12s'),
    ('flag', UINT2),
    ('x_scale', SINT4),
    ('y_scale', SINT4),
    ('z_scale', SINT4),
    ('x_size', SINT4),
    ('y_size', SINT4),
    ('z_size', SINT4),
    ('x_location', SINT4),
    ('y_location', SINT4),
    ('z_location', SINT4),
    ('maximum_range', SINT4),
    ('data_type', UINT2),
    ('projection_name', '12s'),
    ('input_data_type', UINT2),
    ('projection_type', UINT1),
    ('spare_1', '1s'),              # 1 bytes
    ('radial_smoother', SINT2),
    ('times_run', SINT2),
    ('zr_constant', SINT4),
    ('zr_exponent', SINT4),
    ('x_smoother', SINT2),
    ('y_smoother', SINT2),
    ('product_specific_bytes', '80s'),  # 80 bytes:
    ('minor_task_suffix', '16s'),
    ('spare_2', '12s'),             # 12 bytes
    ('color_scale_def', '48s')      # 48 bytes: color_scale_def
)

# 12 bytes: ymds_time Structure (section 4.2.76, page 72)
YMDS_TIME = (
    ('seconds', SINT4),
    ('milliseconds', UINT2),    # milliseconds in lowest 10 bits,
    ('year', SINT2),
    ('month', SINT2),
    ('day', SINT2),
)

# 48 bytes: color_scale_def (section 4.2.5, page 34)
COLOR_SCALE_DEF = (
    ('iflags', UINT4),
    ('istart', SINT4),
    ('istep', SINT4),
    ('icolcnt', SINT2),
    ('iset_and_scale', UINT2),
    ('ilevel_seams', '32s')     # 32 bytes: UINT2[16]
)

# 308 bytes : product_end (section 4.2.24)
PRODUCT_END = (
    ('site_name', '16s'),
    ('iris_version_created', '8s'),
    ('ingest_iris_version', '8s'),
    ('ingest_time', '12s'),         # 12 bytes: ymds_time
    ('spare_0', '28s'),             # 28 bytes
    ('GMT_minute_offset_local', SINT2),
    ('ingest_hardware_name_', '16s'),
    ('ingest_site_name_', '16s'),
    ('GMT_minute_offset_standard', SINT2),
    ('latitude', BIN4),
    ('longitude', BIN4),
    ('ground_height', SINT2),
    ('radar_height', SINT2),
    ('prf', SINT4),
    ('pulse_width', SINT4),
    ('signal_processor_type', UINT2),
    ('trigger_rate', UINT2),
    ('samples_used', SINT2),
    ('clutter_filter', '12s'),
    ('number_linear_filter', UINT2),
    ('wavelength', SINT4),
    ('truncation_height', SINT4),
    ('first_bin_range', SINT4),
    ('last_bin_range', SINT4),
    ('number_bins', SINT4),
    ('flag', UINT2),
    ('number_ingest', SINT2),
    ('polarization', UINT2),
    ('horizontal_calibration_i0', SINT2),
    ('horizontal_calibration_noise', SINT2),
    ('horizontal_radar_constant', SINT2),
    ('reciever_bandwidth', UINT2),
    ('horizontal_current_noise', SINT2),
    ('vertical_current_noise', SINT2),
    ('ldr_offset', SINT2),
    ('zdr_offset', SINT2),
    ('tcf_cal_flags_1', UINT16_T),
    ('tcf_cal_flags_2', UINT16_T),
    ('spare_1', '18s'),             # 18 bytes
    ('standard_parallel_1', BIN4),
    ('standard_parallel_2', BIN4),
    ('earth_radius', UINT4),
    ('inverse_flatting', UINT4),
    ('fault_status', UINT4),
    ('input_mask', UINT4),
    ('number_log_filter', UINT2),
    ('cluttermap', UINT2),
    ('latitude_projection', BIN4),
    ('longitude_projection', BIN4),
    ('product_sequence_number', SINT2),
    ('spare_2', '32s'),             # 32 bytes
    ('melting_level', SINT2),
    ('radar_height_above_reference', SINT2),
    ('number_elements', SINT2),
    ('mean_wind_speed', UINT1),
    ('mean_wind_direction', BIN1),
    ('spare_3', '2s'),              # 2 bytes
    ('tz_name', '8s'),
    ('extended_product_header_offset', UINT4),
    ('spare_4', '4s'),              # 4 bytes
)

# 4884 bytes ingest_header Structure (section 4.2.16, page 40)
INGEST_HEADER = (
    ('structure_header', '12s'),        # 12 bytes: structure_header
    ('ingest_configuration', '480s'),   # 480 bytes: ingest_configuration
    ('task_configuration', '2612s'),    # 2612 bytes: task_configuration
    ('spare_0', '732s'),                # 732 bytes
    ('gparm', '128s'),                  # 128 bytes
    ('reserved', '920s'),               # 920 bytes
)

# 480 bytes ingest_configuration Structure (section 4.2.14, page 38)
INGEST_CONFIGURATION = (
    ('filename', '80s'),
    ('number_files', SINT2),
    ('number_sweeps_completed', SINT2),
    ('total_size', SINT4),
    ('volume_scan_start_time', '12s'),  # 12 bytes: ymds_time
    ('spare_0', '12s'),                 # 12 bytes
    ('ray_header_bytes', SINT2),
    ('extended_ray_header_bytes', SINT2),
    ('number_task_config_table', SINT2),
    ('playback_version', SINT2),
    ('spare_1', '4s'),                  # 4 bytes
    ('iris_version', '8s'),
    ('hardware_site', '16s'),
    ('gmt_offset_minutes_local', SINT2),
    ('site_name', '16s'),
    ('gmt_offset_minutes_standard', SINT2),
    ('latitude_radar', BIN4),
    ('longitude_radar', BIN4),
    ('height_site', SINT2),
    ('height_radar', SINT2),
    ('resolution_rays', UINT2),
    ('first_ray_index', UINT2),
    ('number_rays_sweep', UINT2),
    ('gparam_bytes', SINT2),
    ('altitude_radar', SINT4),
    ('velocity_east', SINT4),
    ('velocity_north', SINT4),
    ('velocity_up', SINT4),
    ('antenna_offset_starboard', SINT4),
    ('antenna_offset_bow', SINT4),
    ('antenna_offset_up', SINT4),
    ('fault_status', UINT4),
    ('melting_layer', SINT2),
    ('spare_2', '2s'),              # 2 bytes
    ('local_timezone', '8s'),
    ('flags', UINT4),
    ('configuration_name', '16s'),
    ('spare_3', '228s')
)

# 2612 bytes: task_configuration Structure (section 4.2.50, page 61)
TASK_CONFIGURATION = (
    ('structure_header', '12s'),    # 12 bytes: structure_header
    ('task_sched_info', '120s'),    # 120 bytes: task_sched_info
    ('task_dsp_info', '320s'),      # 320 bytes: task_dsp_info
    ('task_calib_info', '320s'),    # 320 bytes: task_calib_info
    ('task_range_info', '160s'),    # 160 bytes: task_range_info
    ('task_scan_info', '320s'),     # 320 bytes: task_scan_info
    ('task_misc_info', '320s'),     # 320 bytes: task_misc_info
    ('task_end_info', '320s'),      # 320 bytes: task_end_info
    ('comments', '720s'),
)

# 120 bytes: task_sched_info Structure (section 4.2.61, page 65)
TASK_SCHED_INFO = (
    ('start_time', SINT4),
    ('stop_time', SINT4),
    ('skip_time', SINT4),
    ('last_run_time', SINT4),
    ('time_used_last_run', SINT4),
    ('last_run_day', SINT4),
    ('flag', UINT2),
    ('spare_0', '94s'),
)

# 320 bytes: task_dsp_info Structure (section 4.2.51, page 61)
TASK_DSP_INFO = (
    ('major_mode', UINT2),
    ('dsp_type', UINT2),
    ('current_data_type_mask', '24s'),      # 24 bytes: dsp_data_mask
    ('original_data_type_mask', '24s'),     # 24 bytes: dsp_data_mask
    ('task_dsp_mode', '32s'),               # 32 bytes: task_dsp_mode_batch
    ('spare_0', '52s'),
    ('prf', SINT4),
    ('pulse_width', SINT4),
    ('multi_prf_flag', UINT2),
    ('dual_prf_delay', SINT2),
    ('agc_feedback_code', UINT2),
    ('sample_size', SINT2),
    ('gain_control_flag', UINT2),
    ('clutter_filter_name', '12s'),
    ('linear_filter_first_bin', UINT1),
    ('log_filter_first_bin', UINT1),
    ('attenuation', SINT2),
    ('gas_attenuation', UINT2),
    ('cluttermap_flag', UINT2),
    ('tranmitter_phase_sequence', UINT2),
    ('ray_header_mask', UINT4),
    ('playback_flag', UINT2),
    ('spare_1', '2s'),
    ('custom_ray_header_name', '16s'),
    ('spare_2', '120s')
)

# 24 bytes: dsp_data_mask Structure (section 4.2.7, page 36)
DSP_DATA_MASK = (
    ('mask_word_0', UINT4),
    ('extended_header_type', UINT4),
    ('mask_word_1', UINT4),
    ('mask_word_2', UINT4),
    ('mask_word_3', UINT4),
    ('mask_word_4', UINT4),
)

# 32 bytes: task_dsp_mode_batch (section 4.2.52, page 62)
TASK_DSP_MODE_BATCH = (
    ('low_prf_hz', UINT2),
    ('low_prf_factional', UINT2),
    ('low_prf_sample_size', SINT2),
    ('low_prf_range_averaging', SINT2),
    ('reflectivity_unfolding_threshold', SINT2),
    ('velocity_unfolding_threshold', SINT2),
    ('width_unfolding_threshold', SINT2),
    ('spare_0', '18s'),
)

# 320 bytes: task_calib_info Structure (section 4.2.49, page 59)
TASK_CALIB_INFO = (
    ('reflectivity_slope', SINT2),
    ('reflectivity_noise_threshold', SINT2),
    ('clutter_correction_threshold', SINT2),
    ('sqi_threshold', SINT2),
    ('power_threshold', SINT2),
    ('spare_0', '8s'),
    ('reflectivity_calibration', SINT2),
    ('uncorrected_reflectivity_threshold_flags', UINT2),
    ('corrected_reflectivity_threshold_flags', UINT2),
    ('velocity_threshold_flags', UINT2),
    ('width_threshold_flags', UINT2),
    ('zdr_threshold_flags', UINT2),
    ('spare_1', '6s'),
    ('flags', UINT2),
    ('spare_2', '2s'),
    ('ldr_bias', SINT2),
    ('zdr_bias', SINT2),
    ('nexrad_clutter_threshold', SINT2),
    ('nexrad_clutter_skip', UINT2),
    ('horizontal_i0_calibration', SINT2),
    ('vertical_i0_calibration', SINT2),
    ('horizontal_noise_calibration', SINT2),
    ('vertical_noise_calibration', SINT2),
    ('horizontal_radar_constant', SINT2),
    ('vertical_radar_constant', SINT2),
    ('reciever_bandwidth', UINT2),
    ('flags2', UINT16_T),
    ('spare_3', '256s'),
)

# 160 bytes: task_range_info Structure (section 4.2.58, page 64)
TASK_RANGE_INFO = (
    ('first_bin_range', SINT4),
    ('last_bin_range', SINT4),
    ('number_input_bins', SINT2),
    ('number_output_bins', SINT2),
    ('step_input_bins', SINT4),
    ('step_output_bins', SINT4),
    ('variable_range_bin_flag', UINT2),
    ('range_bin_averaging_flag', SINT2),
    ('spare_0', '136s'),
)

# 320 bytes: task_scan_info Structure (section 4.2.60, page 65)
TASK_SCAN_INFO = (
    ('antenna_scan_mode', UINT2),
    ('angular_resolution_desired', SINT2),
    ('spare_0', '2s'),
    ('number_sweeps', SINT2),
    ('task_scan_type_scan_info', '200s'),   # 200 bytes: task_foo_scan_info
    ('spare_1', '112s'),
)

# 200 bytes: task_rhi_scan_info Structure (section 4.2.59, page 64)
TASK_RHI_SCAN_INFO = (
    ('lower_elevation_limit', UINT2),
    ('upper_elevation_limit', UINT2),
    ('azimuth_list', '80s'),            # UINT2[40]
    ('spare_0', '115s'),
    ('start_first_sector_flag', 'c'),   # unknown type
)

# 200 bytes: task_ppi_scan_info (section 4.2.57, page 64)
TASK_PPI_SCAN_INFO = (
    ('left_azimuth_limit', BIN2),
    ('right_azimuth_limit', BIN2),
    ('elevation_list', '80s'),          # UINT2[40]
    ('spare_0', '115s'),
    ('start_first_section_flag', 'c'),  # unknown type
)

# 200 bytes: task_file_scan_info (section 4.2.54, page 63)
TASK_FILE_SCAN_INFO = (
    ('first_azimuth', UINT2),
    ('first_elevation', UINT2),
    ('filename', '12s'),
    ('spare_0', '184s'),
)

# 200 bytes: task_manual_scan_info (section 4.2.55, page 63)
TASK_MANUAL_SCAN_INFO = (
    ('flags', UINT2),
    ('spare_0', '198s'),
)

# 320 bytes: task_misc_info Structure (section 4.2.55, page 63)
TASK_MISC_INFO = (
    ('wavelength', SINT4),
    ('tr_serial_number', '16s'),
    ('transmit_power', SINT4),
    ('flags', UINT2),
    ('polarization_type', UINT2),
    ('trucation_height', SINT4),
    ('spare_0', '18s'),             # 18 bytes
    ('spare_1', '12s'),             # 12 bytes
    ('comment_bytes', SINT2),
    ('horizontal_beamwidth', BIN4),
    ('vertical_beamwidth', BIN4),
    ('customer_storage', '40s'),    # 40 bytes, uint4[10]
    ('spare_2', '208s'),
)

# 320 bytes: task_end_info Structure (section 4.2.53, page 62)
TASK_END_INFO = (
    ('task_major_number', SINT2),
    ('task_minor_number', SINT2),
    ('task_configuration_file_name', '12s'),
    ('task_description', '80s'),
    ('number_tasks', SINT4),
    ('task_state', UINT2),
    ('spare_0', '2s'),
    ('task_data_time', '12s'),      # 12 bytes: ymds_time
    ('spare_1', '204s'),
)


# 12 bytes raw_prod_bhdr structure (section 4.2.30, page 50)
RAW_PROD_BHDR = (
    ('record_number', SINT2),
    ('sweep_number', SINT2),
    ('first_ray_offset', SINT2),
    ('ray_number', SINT2),
    ('flags', UINT2),
    ('spare_0', '2s'),
)

# 76 bytes ingest_data_header (section 4.2.15, pages 40)
INGEST_DATA_HEADER = (
    ('structure_header', '12s'),    # 12 bytes: structure_header
    ('sweep_start_time', '12s'),    # 12 bytes: ymds_time
    ('sweep_number', SINT2),
    ('number_rays_sweep', SINT2),
    ('first_ray_index', SINT2),
    ('number_rays_file_expected', SINT2),
    ('number_rays_file_actual', SINT2),
    ('fixed_angle', BIN2),
    ('bit_per_bin', SINT2),
    ('data_type', UINT2),
    ('spare_0', '36s')      # 36 bytes
)



"""
Cython wrapper around University of Washington 4DD code.
"""

cimport pyart.io._rsl_h as _rsl_h

cdef extern from "dealias_fourdd.h":

    int dealias_fourdd(_rsl_h.Volume* rvVolume, _rsl_h.Volume* soundVolume,
                       _rsl_h.Volume* lastVolume, float missingVal,
                       float compthresh, float compthresh2, float thresh,
                       float ckval, float stdthresh, float epsilon, 
                       int maxcount, int pass2, int rm, int proximity,
                       int mingood, int filt, int ba_mincount, 
                       int ba_edgecount)


cdef extern from "sounding_to_volume.h":
    
    int sounding_to_volume(_rsl_h.Volume* soundVolume, float missingVal,
                           float *height_array, float *speed_array,
                           float *direction_array, int nlevels, 
                           float maxshear, int sign)



#! /usr/bin/env python
"""
Make a small NEXRAD (WSR-88D) Level II file with compressed records.

The file created by this script only contains the first 254 records
(120 radials) of the full sized file.  It should not be used as a full sized
example, only for testing the decompression routines.
"""

f = open('Level2_KATX_20130717_1950.ar2v', 'rb')
o = open('example_nexrad_archive_msg31_compressed.ar2v', 'wb')

o.write(f.read(24))             # volume header
o.write(f.read(12527 + 4))      # first series of compressed records
o.write(f.read(105727 + 4))     # second series of compressed records
o.close()



""" Unit Tests for Py-ART's correct/proc_phase.py module. """

# Can also be run as a script to create a ray_plot.png file
# python test_phase_proc.py
# This file should look like the reference_ray_plot.png file

# Use:
# python test_phase_proc.py -r
# to recreate the reference_rays.npz and reference_ray_plot.png files

import os
import warnings

import numpy as np
import pytest

try:
    import cvxopt
    cvxopt_available = True
except ImportError:
    cvxopt_available = False

try:
    import glpk
    glpk_available = True
except ImportError:
    glpk_available = False


try:
    import cylp.cy
    cylp_available = True
except ImportError:
    cylp_available = False

import pyart


PATH = os.path.dirname(__file__)
REFERENCE_RAYS_FILE = os.path.join(PATH, 'reference_rays.npz')


@pytest.mark.skipif(not glpk_available,
                    reason="GLPK is not installed.")
def test_phase_proc_lp_glpk():
    radar, phidp, kdp = perform_phase_processing()
    ref = np.load(REFERENCE_RAYS_FILE)
    assert _ratio(ref['reference_phidp'], phidp['data']) <= 0.01
    assert _ratio(ref['reference_kdp'], kdp['data']) <= 0.01
    assert _ratio(ref['reference_unfolded_phidp'],
                  radar.fields['unfolded_differential_phase']['data']) <= 0.01


@pytest.mark.skipif(not cvxopt_available,
                    reason="CVXOPT is not installed.")
def test_phase_proc_lp_cvxopt():
    from cvxopt import solvers
    solvers.options['LPX_K_MSGLEV'] = 0     # supress screen output
    radar, phidp, kdp = perform_phase_processing('cvxopt')
    ref = np.load(REFERENCE_RAYS_FILE)
    assert _ratio(ref['reference_phidp'], phidp['data']) <= 0.01
    assert _ratio(ref['reference_kdp'], kdp['data']) <= 0.01
    assert _ratio(ref['reference_unfolded_phidp'],
                  radar.fields['unfolded_differential_phase']['data']) <= 0.01


@pytest.mark.skipif(not cylp_available,
                    reason="CyLP is not installed.")
def test_phase_proc_lp_cylp():
    with warnings.catch_warnings():
        # ignore FutureWarnings as CyLP emits a number of these
        warnings.simplefilter("ignore", category=FutureWarning)
        radar, phidp, kdp = perform_phase_processing('cylp')
    ref = np.load(REFERENCE_RAYS_FILE)
    assert _ratio(ref['reference_phidp'], phidp['data']) <= 0.01
    assert _ratio(ref['reference_kdp'], kdp['data']) <= 0.01
    assert _ratio(ref['reference_unfolded_phidp'],
                  radar.fields['unfolded_differential_phase']['data']) <= 0.01

@pytest.mark.skipif(not cylp_available,
                    reason="CyLP is not installed.")
def test_phase_proc_lp_gf_cylp():
    with warnings.catch_warnings():
        # ignore FutureWarnings as CyLP emits a number of these
        warnings.simplefilter("ignore", category=FutureWarning)
        radar, phidp, kdp = perform_phase_processing_gf('cylp')
    ref = np.load(REFERENCE_RAYS_FILE)
    assert _ratio(ref['reference_phidp'], phidp['data']) <= 0.05
    assert _ratio(ref['reference_kdp'], kdp['data']) <= 0.05
    assert _ratio(ref['reference_unfolded_phidp'],
                  radar.fields['unfolded_differential_phase']['data']) <= 0.05


@pytest.mark.skipif(not cylp_available,
                    reason="CyLP is not installed.")
def test_phase_proc_lp_cylp_mp():
    with warnings.catch_warnings():
        # ignore FutureWarnings as CyLP emits a number of these
        warnings.simplefilter("ignore", category=FutureWarning)
        radar, phidp, kdp = perform_phase_processing('cylp_mp')
    ref = np.load(REFERENCE_RAYS_FILE)
    assert _ratio(ref['reference_phidp'], phidp['data']) <= 0.01
    assert _ratio(ref['reference_kdp'], kdp['data']) <= 0.01
    assert _ratio(ref['reference_unfolded_phidp'],
                  radar.fields['unfolded_differential_phase']['data']) <= 0.01


def _ratio(a1, a2):
    """ Ratio the sum of the abs difference vs sum abs of two vectors. """
    abs_residues = np.abs(a1 - a2).sum()
    avg_abs_sum = 0.5 * np.abs(a1).sum() + 0.5 * np.abs(a2).sum()
    return abs_residues / avg_abs_sum


def perform_phase_processing(LP_solver='pyglpk'):
    """ Perform LP phase processing on a single ray radar. """
    radar = pyart.testing.make_single_ray_radar()
    phidp, kdp = pyart.correct.phase_proc_lp(radar, 0.0, LP_solver=LP_solver)
    return radar, phidp, kdp

def perform_phase_processing_gf(LP_solver='pyglpk'):
    """ Perform LP phase processing on a single ray radar. """
    radar = pyart.testing.make_single_ray_radar()
    my_gatefilter = pyart.filters.GateFilter(radar)
    my_gatefilter.exclude_below('normalized_coherent_power', 0.5)
    my_gatefilter.exclude_below('cross_correlation_ratio', 0.8)
    phidp, kdp = pyart.correct.phase_proc_lp_gf(radar, gatefilter=my_gatefilter,
                                                LP_solver=LP_solver, doc=15,
                                                ncpts=20, system_phase=-140.1)
    return radar, phidp, kdp


def save_reference_rays(radar, phidp, kdp):
    """ Save the phase processed rays to REFERENCE_RAY_FILE. """
    unfolded = radar.fields['unfolded_differential_phase']
    np.savez(
        REFERENCE_RAYS_FILE,
        reference_phidp=phidp['data'],
        reference_kdp=kdp['data'],
        reference_unfolded_phidp=unfolded['data'])


def make_plot(range_km, unfolded_phidp, refl, phidp, kdp, filename):
    """
    Make and save a plot of PhiDP, reflectivity and KDP.

    Parameters
    ----------
    range_km : array
        Gate range in kilometers.
    unfolded_phidp, refl, phidp, kdp : dict
        Dictionary containing the unfolded phiDP, reflectivity,
        LP filtered phiDP and KDP in the data key
    filename :
        Filename to save the file to.

    """

    from matplotlib import pyplot as plt
    fig = plt.figure(figsize=[10, 5])
    ax = fig.add_subplot(111)

    # filtered phidp and unfolded phidp
    p1, = ax.plot(range_km, phidp['data'][0], 'b-')
    p2, = ax.plot(range_km, unfolded_phidp['data'][0], 'g-')

    # set labels
    ax.set_ylim(0, 250)
    ax.set_ylabel('Differential phase shift (degrees)')
    ax.set_xlabel('Range (km)')

    # plot KDP and reflectivity on second axis
    ax2 = ax.twinx()
    p3, = ax2.plot(range_km, kdp['data'][0], 'r-')
    p4, = ax2.plot(range_km, refl['data'][0]/10.)

    # decorate and save
    ax2.yaxis.grid(color='gray', linestyle='dashed')
    ax.legend([p1, p2, p3, p4],
              ["Filtered phiDP", "Unfolded phiDP", 'KDP', 'Z/10.0'],
              loc='upper left')
    fig.savefig(filename)


if __name__ == "__main__":
    import sys

    radar, phidp, kdp = perform_phase_processing()
    filename = 'ray_plot.png'
    if sys.argv[-1] == '-r':    # regenerate reference files
        save_reference_rays(radar, phidp, kdp)
        filename = 'reference_ray_plot.png'

    make_plot(radar.range['data'] / 1000.0,
              radar.fields['unfolded_differential_phase'],
              radar.fields['reflectivity'],
              phidp, kdp, filename)



""" Unit Tests for Py-ART's io/nexrad_level2.py module. """

import datetime
import bz2
from io import BytesIO

import numpy as np
from numpy.testing import assert_array_equal
from numpy.testing import assert_almost_equal
import pytest

import pyart
import pyart.io.nexrad_level2 as nexrad_level2

# create a NEXRADLevel2File from a UNCOMPRESSED dummy file
# pyart/testing/data/example_nexrad_archive.bz2
UNCOMPRESSED_FILE = bz2.BZ2File(pyart.testing.NEXRAD_ARCHIVE_MSG31_FILE, 'rb')
nfile = nexrad_level2.NEXRADLevel2File(UNCOMPRESSED_FILE)
nfile.close()


# attributes
def test_radial_messages():
    assert len(nfile.radial_records) == 7200


def test_nscans():
    assert nfile.nscans == 16


def test_scan_msgs():
    assert len(nfile.scan_msgs) == 16
    assert len(nfile.scan_msgs[0]) == 720
    assert len(nfile.scan_msgs[-1]) == 360
    assert nfile.scan_msgs[0][10] == 10
    assert nfile.scan_msgs[10][10] == 5050


def test_volume_header():
    assert nfile.volume_header['date'] == 15904
    assert nfile.volume_header['extension'] == b'501'
    assert nfile.volume_header['icao'] == b'KATX'
    assert nfile.volume_header['tape'] == b'AR2V0006.'
    assert nfile.volume_header['time'] == 71424000


# methods
def test_get_azimuth_angles():
    angles = nfile.get_azimuth_angles([0, 1])
    assert angles.shape == (1440, )
    assert round(angles[0]) == 350.0
    assert round(angles[10]) == 355.0


def test_get_data_ref():
    data = nfile.get_data('REF', 1832, [0, 1])
    assert data.shape == (1440, 1832)
    assert round(data[0, 0]) == -32.0
    assert np.all(data == -32)
    assert np.ma.is_masked(data[-1, -1])


def test_get_data_vel():
    data = nfile.get_data('VEL', 1192, [1, 3])
    assert data.shape == (1440, 1192)
    assert round(data[0, 0], 2) == -63.5
    assert np.all(data == -63.5)


def test_get_data_sw():
    data = nfile.get_data('SW', 1192, [1, 3])
    assert data.shape == (1440, 1192)
    assert round(data[0, 0], 2) == -63.5
    assert np.all(data == -63.5)


def test_get_data_zdr():
    data = nfile.get_data('ZDR', 1192, [0, 2])
    assert data.shape == (1440, 1192)
    assert round(data[0, 0], 3) == -7.875
    assert np.all(data == -7.875)


def test_get_data_phi():
    data = nfile.get_data('PHI', 1192, [0, 2])
    assert data.shape == (1440, 1192)
    assert round(data[0, 0], 1) == 180.5


def test_get_data_rho():
    data = nfile.get_data('RHO', 1192, [0, 2])
    assert data.shape == (1440, 1192)
    assert_almost_equal(data[0, 0], 0.208, 3)


def test_get_data_ref_raw():
    data = nfile.get_data('REF', 1832, [0, 1], True)
    assert data.shape == (1440, 1832)
    assert round(data[0, 0]) == 2
    assert np.all(data[:720, :] == 2)
    assert np.all(data[720:, :1192] == 2)
    assert data[-1, -1] == 1        # folded


def test_get_data_vel_raw():
    data = nfile.get_data('VEL', 1192, [1, 3], True)
    assert data.shape == (1440, 1192)
    assert round(data[0, 0], 2) == 2
    assert np.all(data == 2)


def test_get_data_sw_raw():
    data = nfile.get_data('SW', 1192, [1, 3], True)
    assert data.shape == (1440, 1192)
    assert round(data[0, 0]) == 2
    assert np.all(data == 2)


def test_get_data_zdr_raw():
    data = nfile.get_data('ZDR', 1192, [0, 2], True)
    assert data.shape == (1440, 1192)
    assert round(data[0, 0]) == 2
    assert np.all(data == 2)


def test_get_data_phi_raw():
    data = nfile.get_data('PHI', 1192, [0, 2], True)
    assert data.shape == (1440, 1192)
    assert round(data[0, 0]) == 514     # big-endiann packing of 0,0
    assert np.all(data == 514)


def test_get_data_rho_raw():
    data = nfile.get_data('RHO', 1192, [0, 2], True)
    assert data.shape == (1440, 1192)
    assert round(data[0, 0]) == 2
    assert np.all(data == 2)


def test_get_data_field_not_present():
    # should return a masked array with all values masked
    data = nfile.get_data('VEL', 1192, [0])
    assert data.shape == (720, 1192)
    assert np.all(data.mask)


def test_get_elevation_angles():
    angles = nfile.get_elevation_angles([0, 1])
    assert angles.shape == (1440, )
    assert round(angles[0], 3) == 0.747
    assert round(angles[10], 3) == 0.670


def test_get_nrays():
    assert nfile.get_nrays(0) == 720
    assert nfile.get_nrays(4) == 360


def test_get_range():
    _range = nfile.get_range(0, 'REF')
    assert_array_equal(_range, 2125 + 250 * np.arange(1832))
    _range = nfile.get_range(4, 'REF')
    assert_array_equal(_range, 2125 + 250 * np.arange(1352))


def test_get_times():
    time_start, times = nfile.get_times([0, 1])
    assert time_start == datetime.datetime(2013, 7, 17, 19, 50, 21)
    assert times.shape == (1440, )
    assert round(times[0], 2) == 0.65
    assert round(times[10], 2) == 0.92


def test_location():
    lat, lon, alt = nfile.location()
    assert round(lat, 1) == 48.2
    assert round(lon, 1) == -122.5
    assert round(alt) == 195


def test_scan_info():
    scan_info = nfile.scan_info()

    assert len(scan_info) == 16

    nrays = [s['nrays'] for s in scan_info]
    ref_nrays = [720, 720, 720, 720, 360, 360, 360, 360, 360, 360, 360, 360,
                 360, 360, 360, 360]
    assert nrays == ref_nrays

    assert scan_info[0]['moments'] == ['REF', 'ZDR', 'PHI', 'RHO']
    assert scan_info[1]['moments'] == ['REF', 'VEL', 'SW']
    assert scan_info[2]['moments'] == ['REF', 'ZDR', 'PHI', 'RHO']
    assert scan_info[3]['moments'] == ['REF', 'VEL', 'SW']
    all_moments = ['REF', 'VEL', 'SW', 'ZDR', 'PHI', 'RHO']
    assert scan_info[4]['moments'] == all_moments
    assert scan_info[5]['moments'] == all_moments
    assert scan_info[6]['moments'] == all_moments
    assert scan_info[7]['moments'] == all_moments
    assert scan_info[8]['moments'] == all_moments
    assert scan_info[9]['moments'] == all_moments
    assert scan_info[10]['moments'] == all_moments
    assert scan_info[11]['moments'] == all_moments
    assert scan_info[12]['moments'] == all_moments
    assert scan_info[13]['moments'] == all_moments
    assert scan_info[14]['moments'] == all_moments
    assert scan_info[15]['moments'] == all_moments

    assert scan_info[0]['ngates'] == [1832, 1192, 1192, 1192]
    assert scan_info[1]['ngates'] == [1192, 1192, 1192]
    assert scan_info[2]['ngates'] == [1676, 1192, 1192, 1192]
    assert scan_info[3]['ngates'] == [1192, 1192, 1192]
    assert scan_info[4]['ngates'] == [1352, 1192, 1192, 1192, 1192, 1192]
    assert scan_info[5]['ngates'] == [1112, 1112, 1112, 1112, 1112, 1112]
    assert scan_info[6]['ngates'] == [940, 940, 940, 940, 940, 940]
    assert scan_info[7]['ngates'] == [800, 800, 800, 800, 800, 800]
    assert scan_info[8]['ngates'] == [704, 704, 704, 704, 704, 704]
    assert scan_info[9]['ngates'] == [540, 540, 540, 540, 540, 540]
    assert scan_info[10]['ngates'] == [500, 500, 500, 500, 500, 500]
    assert scan_info[11]['ngates'] == [460, 460, 460, 460, 460, 460]
    assert scan_info[12]['ngates'] == [388, 388, 388, 388, 388, 388]
    assert scan_info[13]['ngates'] == [332, 332, 332, 332, 332, 332]
    assert scan_info[14]['ngates'] == [280, 280, 280, 280, 280, 280]
    assert scan_info[15]['ngates'] == [240, 240, 240, 240, 240, 240]


# create a NEXRADLevel2File from a COMPRESSED file
# pyart/testing/data/example_nexrad_archive_compressed.ar2v
COMPRESSED_FILE = pyart.testing.NEXRAD_ARCHIVE_MSG31_COMPRESSED_FILE
cfile = nexrad_level2.NEXRADLevel2File(COMPRESSED_FILE)


def test_compressed_attributes():
    # the compressed archive only contains the first 120 radials
    assert len(cfile.radial_records) == 120
    assert cfile.nscans == 1
    assert len(cfile.scan_msgs) == 1
    assert len(cfile.scan_msgs[0]) == 120
    assert cfile.volume_header['date'] == 15904
    assert cfile.volume_header['extension'] == b'501'
    assert cfile.volume_header['icao'] == b'KATX'
    assert cfile.volume_header['tape'] == b'AR2V0006.'
    assert cfile.volume_header['time'] == 71424000


# methods
def test_compressed_get_azimuth_angles():
    angles = cfile.get_azimuth_angles([0])
    assert angles.shape == (120, )
    assert round(angles[0]) == 350.0
    assert round(angles[10]) == 355.0


def test_compressed_get_data():
    data = cfile.get_data('REF', 1832, [0])
    assert data.shape == (120, 1832)
    assert round(data[0, 0], 1) == 10.5

    data = cfile.get_data('PHI', 1192, [0])
    assert data.shape == (120, 1192)
    assert round(data[0, 0]) == 254

    data = cfile.get_data('RHO', 1192, [0])
    assert data.shape == (120, 1192)
    assert_almost_equal(data[0, 0], 0.688, 3)

    data = cfile.get_data('ZDR', 1192, [0])
    assert data.shape == (120, 1192)
    assert round(data[0, 0], 3) == -7.875


def test_compressed_get_data_raw():
    data = cfile.get_data('REF', 1832, [0], True)
    assert data.shape == (120, 1832)
    assert round(data[0, 0]) == 87

    data = cfile.get_data('PHI', 1192, [0], True)
    assert data.shape == (120, 1192)
    assert round(data[0, 0]) == 722

    data = cfile.get_data('RHO', 1192, [0], True)
    assert data.shape == (120, 1192)
    assert round(data[0, 0]) == 146

    data = cfile.get_data('ZDR', 1192, [0], True)
    assert data.shape == (120, 1192)
    assert round(data[0, 0]) == 2


def test_compressed_get_elevation_angles():
    angles = cfile.get_elevation_angles([0])
    assert angles.shape == (120, )
    assert round(angles[0], 3) == 0.747
    assert round(angles[10], 3) == 0.670


def test_compressed_get_nrays():
    assert cfile.get_nrays(0) == 120


def test_compressed_get_range():
    _range = cfile.get_range(0, 'REF')
    assert_array_equal(_range, 2125 + 250 * np.arange(1832))


def test_compressed_get_times():
    time_start, times = cfile.get_times([0])
    assert time_start == datetime.datetime(2013, 7, 17, 19, 50, 21)
    assert times.shape == (120, )
    assert round(times[0], 2) == 0.65
    assert round(times[10], 2) == 0.92


def test_compressed_location():
    lat, lon, alt = cfile.location()
    assert round(lat, 1) == 48.2
    assert round(lon, 1) == -122.5
    assert round(alt) == 195


def test_compressed_scan_info():
    scan_info = cfile.scan_info()

    assert len(scan_info) == 1
    assert scan_info[0]['moments'] == ['REF', 'ZDR', 'PHI', 'RHO']
    assert scan_info[0]['ngates'] == [1832, 1192, 1192, 1192]
    assert scan_info[0]['nrays'] == 120


def test_bad_compression_header():

    # read the beginning of the compressed file
    f = open(COMPRESSED_FILE, 'rb')
    head = f.read(36)
    f.close()

    # corrupt the compression header and write to in memory file
    head = head[:28] + b'XX' + head[30:]
    corrupt_file = BytesIO()
    corrupt_file.write(head)
    corrupt_file.seek(0)

    # should raise IOError
    pytest.raises(IOError, nexrad_level2.NEXRADLevel2File, corrupt_file)


def test_msg1_missing_location():
    # Message 1 files do not contain location information
    uncompressed_file = bz2.BZ2File(
        pyart.testing.NEXRAD_ARCHIVE_MSG1_FILE, 'rb')
    nfile = nexrad_level2.NEXRADLevel2File(uncompressed_file)

    latitude, longitude, height = nfile.location()
    assert latitude == 0.0
    assert longitude == 0.0
    assert height == 0.0


def test_invalid_msg31_block():
    # This should never happen with real NEXRAD files...
    block_name, dic = nexrad_level2._get_msg31_data_block(b'aaa', 0)
    assert len(dic) == 0


def test_broken_file():
    uncompressed_file = bz2.BZ2File(
        pyart.testing.NEXRAD_ARCHIVE_MSG1_FILE, 'rb')

    broken_string = uncompressed_file.read(202)
    broken_file = BytesIO()
    broken_file.write(broken_string)
    broken_file.seek(0)

    # should raise ValueError
    pytest.raises(ValueError, nexrad_level2.NEXRADLevel2File, broken_file)


def test_1ms_vel_message():
    # read in a velocity message from a MSG1 file
    uncompressed_file = bz2.BZ2File(
        pyart.testing.NEXRAD_ARCHIVE_MSG1_FILE, 'rb')
    # seek pack the volume header, compression header and reflectivity data
    uncompressed_file.seek(24 + 12 + 897408)
    buf = uncompressed_file.read(2432)  # read a record

    # create a fake message with a 1 m/s velocity resolution
    fake_buf = buf[:16+43] + b'\x04' + buf[16+44:]

    # check the velocity scale
    new_pos, dic = nexrad_level2._get_record_from_buf(fake_buf, 0)
    assert dic['VEL']['scale'] == 1.0



""" Unit Tests for Py-ART's retrieve/echo_class.py module. """

import numpy as np
import pytest

import pyart


def test_steiner_conv_strat():
    grid = pyart.testing.make_storm_grid()
    eclass = pyart.retrieve.steiner_conv_strat(grid)
    assert np.all(eclass['data'][25] == np.array(
        [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
         2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0]))



"""
pyart.correct.filters
=====================

Functions for creating gate filters (masks) which can be used it various
corrections routines in Py-ART.

.. autosummary::
    :toctree: generated/

    moment_based_gate_filter
    moment_and_texture_based_gate_filter
    birds_gate_filter
    snr_based_gate_filter
    class_based_gate_filter
    visibility_based_gate_filter
    temp_based_gate_filter
    iso0_based_gate_filter

.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

    GateFilter

"""

from copy import deepcopy

import numpy as np

from ..config import get_field_name, get_metadata
from ..util import texture_along_ray


def moment_based_gate_filter(
        radar, ncp_field=None, rhv_field=None, refl_field=None,
        min_ncp=0.5, min_rhv=None, min_refl=-20., max_refl=100.0):
    """
    Create a filter which removes undesired gates based on moments.

    Creates a gate filter in which the following gates are excluded:

    * Gates where the instrument is transitioning between sweeps.
    * Gates where the reflectivity is outside the interval min_refl, max_refl.
    * Gates where the normalized coherent power is below min_ncp.
    * Gates where the cross correlation ratio is below min_rhi. Using the
      default parameter this filtering is disabled.
    * Gates where any of the above three fields are masked or contain
      invalid values (NaNs or infs).
    * If any of these three fields do not exist in the radar that fields filter
      criteria is not applied.

    Parameters
    ----------
    radar : Radar
        Radar object from which the gate filter will be built.
    refl_field, ncp_field, rhv_field : str
        Names of the radar fields which contain the reflectivity, normalized
        coherent power (signal quality index) and cross correlation ratio
        (RhoHV) from which the gate filter will be created using the above
        criteria. A value of None for any of these parameters will use the
        default field name as defined in the Py-ART configuration file.
    min_ncp, min_rhv : float
        Minimum values for the normalized coherence power and cross
        correlation ratio. Gates in these fields below these limits as well as
        gates which are masked or contain invalid values will be excluded and
        not used in calculation which use the filter. A value of None will
        disable filtering based upon the given field including removing
        masked or gates with an invalid value. To disable the thresholding
        but retain the masked and invalid filter set the parameter to a value
        below the lowest value in the field.
    min_refl, max_refl : float
        Minimum and maximum values for the reflectivity. Gates outside
        of this interval as well as gates which are masked or contain invalid
        values will be excluded and not used in calculation which use this
        filter. A value or None for one of these parameters will disable the
        minimum or maximum filtering but retain the other. A value of None
        for both of these values will disable all filtering based upon the
        reflectivity including removing masked or gates with an invalid value.
        To disable the interval filtering but retain the masked and invalid
        filter set the parameters to values above and below the lowest and
        greatest values in the reflectivity field.

    Returns
    -------
    gatefilter : :py:class:`GateFilter`
        A gate filter based upon the described criteria. This can be
        used as a gatefilter parameter to various functions in pyart.correct.

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if ncp_field is None:
        ncp_field = get_field_name('normalized_coherent_power')
    if rhv_field is None:
        rhv_field = get_field_name('cross_correlation_ratio')

    # filter gates based upon field parameters
    gatefilter = GateFilter(radar)
    gatefilter.exclude_transition()
    if (min_ncp is not None) and (ncp_field in radar.fields):
        gatefilter.exclude_below(ncp_field, min_ncp)
        gatefilter.exclude_masked(ncp_field)
        gatefilter.exclude_invalid(ncp_field)
    if (min_rhv is not None) and (rhv_field in radar.fields):
        gatefilter.exclude_below(rhv_field, min_rhv)
        gatefilter.exclude_masked(rhv_field)
        gatefilter.exclude_invalid(rhv_field)
    if refl_field in radar.fields:
        if min_refl is not None:
            gatefilter.exclude_below(refl_field, min_refl)
            gatefilter.exclude_masked(refl_field)
            gatefilter.exclude_invalid(refl_field)
        if max_refl is not None:
            gatefilter.exclude_above(refl_field, max_refl)
            gatefilter.exclude_masked(refl_field)
            gatefilter.exclude_invalid(refl_field)
    return gatefilter


def moment_and_texture_based_gate_filter(
        radar, zdr_field=None, rhv_field=None, phi_field=None, refl_field=None,
        textzdr_field=None, textrhv_field=None, textphi_field=None,
        textrefl_field=None, wind_size=7, max_textphi=20., max_textrhv=0.3,
        max_textzdr=2.85, max_textrefl=8., min_rhv=0.6):
    """
    Create a filter which removes undesired gates based on texture of moments.

    Creates a gate filter in which the following gates are excluded:
    * Gates where the instrument is transitioning between sweeps.
    * Gates where RhoHV is below min_rhv
    * Gates where the PhiDP texture is above max_textphi.
    * Gates where the RhoHV texture is above max_textrhv.
    * Gates where the ZDR texture is above max_textzdr
    * Gates where the reflectivity texture is above max_textrefl
    * If any of the thresholds is not set or the field (RhoHV, ZDR, PhiDP,
    reflectivity) do not exist in the radar the filter is not applied.

    Parameters
    ----------
    radar : Radar
        Radar object from which the gate filter will be built.
    zdr_field, rhv_field, phi_field, refl_field : str
        Names of the radar fields which contain the differential reflectivity,
        cross correlation ratio, differential phase and reflectivity from
        which the textures will be computed. A value of None for any of these
        parameters will use the default field name as defined in the Py-ART
        configuration file.
    textzdr_field, textrhv_field, textphi_field, textrefl_field : str
        Names of the radar fields given to the texture of the
        differential reflectivity, texture of the cross correlation ratio,
        texture of differential phase and texture of reflectivity. A value
        of None for any of these parameters will use the default field name
        as defined in the Py-ART configuration file.
    wind_size : int
        Size of the moving window used to compute the ray texture.
    max_textphi, max_textrhv, max_textzdr, max_textrefl : float
        Maximum value for the texture of the differential phase, texture of
        RhoHV, texture of Zdr and texture of reflectivity. Gates in these
        fields above these limits as well as gates which are masked or contain
        invalid values will be excluded and not used in calculation which use
        the filter. A value of None will disable filtering based upon the
        given field including removing masked or gates with an invalid value.
        To disable the thresholding but retain the masked and invalid filter
        set the parameter to a value above the highest value in the field.
    min_rhv : float
        Minimum value for the RhoHV. Gates below this limits as well as gates
        which are masked or contain invalid values will be excluded and not
        used in calculation which use the filter. A value of None will disable
        filtering based upon the given field including removing masked or
        gates with an invalid value. To disable the thresholding but retain
        the masked and invalid filter set the parameter to a value below the
        lowest value in the field.

    Returns
    -------
    gatefilter : :py:class:`GateFilter`
        A gate filter based upon the described criteria. This can be
        used as a gatefilter parameter to various functions in pyart.correct.

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if zdr_field is None:
        zdr_field = get_field_name('differential_reflectivity')
    if rhv_field is None:
        rhv_field = get_field_name('cross_correlation_ratio')
    if phi_field is None:
        phi_field = get_field_name('uncorrected_differential_phase')
    if textrefl_field is None:
        textrefl_field = get_field_name('reflectivity_texture')
    if textzdr_field is None:
        textzdr_field = get_field_name('differential_reflectivity_texture')
    if textrhv_field is None:
        textrhv_field = get_field_name('cross_correlation_ratio_texture')
    if textphi_field is None:
        textphi_field = get_field_name('differential_phase_texture')

    # make deepcopy of input radar (we do not want to modify the original)
    radar_aux = deepcopy(radar)

    # compute the textures of the moments and add them into radar object
    if (max_textphi is not None) and (phi_field in radar_aux.fields):
        textphi = texture_along_ray(radar_aux, phi_field, wind_size=wind_size)
        tphi = get_metadata(textphi_field)
        tphi['data'] = textphi
        radar_aux.add_field(textphi_field, tphi)

    if (max_textrhv is not None) and (rhv_field in radar_aux.fields):
        textrho = texture_along_ray(radar_aux, rhv_field, wind_size=wind_size)
        trhv = get_metadata(textrhv_field)
        trhv['data'] = textrho
        radar_aux.add_field(textrhv_field, trhv)

    if (max_textzdr is not None) and (zdr_field in radar_aux.fields):
        textzdr = texture_along_ray(radar_aux, zdr_field, wind_size=wind_size)
        tzdr = get_metadata(textzdr_field)
        tzdr['data'] = textzdr
        radar_aux.add_field(textzdr_field, tzdr)

    if (max_textrefl is not None) and (refl_field in radar_aux.fields):
        textrefl = texture_along_ray(radar_aux, refl_field,
                                     wind_size=wind_size)
        trefl = get_metadata(textrefl_field)
        trefl['data'] = textrefl
        radar_aux.add_field(textrefl_field, trefl)

    # filter gates based upon field parameters
    gatefilter = GateFilter(radar_aux)
    gatefilter.exclude_transition()
    if (min_rhv is not None) and (rhv_field in radar_aux.fields):
        gatefilter.exclude_below(rhv_field, min_rhv)
        gatefilter.exclude_masked(rhv_field)
        gatefilter.exclude_invalid(rhv_field)
    if (max_textphi is not None) and (textphi_field in radar_aux.fields):
        gatefilter.exclude_above(textphi_field, max_textphi)
        gatefilter.exclude_masked(textphi_field)
        gatefilter.exclude_invalid(textphi_field)
    if (max_textrhv is not None) and (textrhv_field in radar_aux.fields):
        gatefilter.exclude_above(textrhv_field, max_textrhv)
        gatefilter.exclude_masked(textrhv_field)
        gatefilter.exclude_invalid(textrhv_field)
    if (max_textzdr is not None) and (textzdr_field in radar_aux.fields):
        gatefilter.exclude_above(textzdr_field, max_textzdr)
        gatefilter.exclude_masked(textzdr_field)
        gatefilter.exclude_invalid(textzdr_field)
    if (max_textrefl is not None) and (textrefl_field in radar_aux.fields):
        gatefilter.exclude_above(textrefl_field, max_textrefl)
        gatefilter.exclude_masked(textrefl_field)
        gatefilter.exclude_invalid(textrefl_field)
    return gatefilter


def birds_gate_filter(
        radar, zdr_field=None, rhv_field=None, refl_field=None,
        vel_field=None, max_zdr=3., max_rhv=0.9, min_refl=0., max_refl=20.,
        vel_lim=1., rmin=2000., rmax=25000., elmin=1., elmax=85.):
    """
    Create a filter which removes data not suspected of being birds

    Creates a gate filter in which the following gates are excluded:

    * Gates where the instrument is transitioning between sweeps.
    * Gates where the reflectivity is beyond min_refl and max_refl
    * Gates where the co-polar correlation coefficient is above max_rhv
    * Gates where the differential reflectivity is above max_zdr
    * Gates where the Doppler velocity is within the interval given by
      +-vel_lim
    * Gates where any of the above fields are masked or contain
      invalid values (NaNs or infs).
    * Gates outside the range given by range min and range max
    * If any of these three fields do not exist in the radar that fields filter
      criteria is not applied.

    Parameters
    ----------
    radar : Radar
        Radar object from which the gate filter will be built.
    refl_field, zdr_field, rhv_field, vel_field : str
        Names of the radar fields which contain the reflectivity, differential
        reflectivity, co-polar correlation coefficient, and Doppler velocity
        from which the gate filter will be created using the above criteria. A
        value of None for any of these parameters will use the default field
        name as defined in the Py-ART configuration file.
    max_zdr, max_rhv : float
        Maximum values for the differential reflectivity and co-polar
        correlation coefficient. Gates in these fields above these limits as
        well as gates which are masked or contain invalid values will be
        excluded and not used in calculation which use the filter.
        A value of None will disable filtering based upon the given field
        including removing masked or gates with an invalid value.
        To disable the thresholding but retain the masked and invalid filter
        set the parameter to a value above the highest value in the field.
    min_refl, max_refl : float
        Minimum and maximum values for the reflectivity.  Gates outside
        of this interval as well as gates which are masked or contain invalid
        values will be excluded and not used in calculation which use this
        filter. A value or None for one of these parameters will disable the
        minimum or maximum filtering but retain the other.  A value of None
        for both of these values will disable all filtering based upon the
        reflectivity including removing masked or gates with an invalid value.
        To disable the interval filtering but retain the masked and invalid
        filter set the parameters to values above and below the lowest and
        greatest values in the reflectivity field.
    rmin, rmax : float
        Minimum and maximum ranges [m]
    elmin, elmax : float
        Minimum and maximum elevations [deg]

    Returns
    -------
    gatefilter : :py:class:`GateFilter`
        A gate filter based upon the described criteria.  This can be
        used as a gatefilter parameter to various functions in pyart.correct.

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if zdr_field is None:
        zdr_field = get_field_name('differential_reflectivity')
    if rhv_field is None:
        rhv_field = get_field_name('cross_correlation_ratio')
    if vel_field is None:
        vel_field = get_field_name('velocity')

    # filter gates based upon field parameters
    gatefilter = GateFilter(radar)
    gatefilter.exclude_transition()
    if (max_zdr is not None) and (zdr_field in radar.fields):
        gatefilter.exclude_above(zdr_field, max_zdr)
        gatefilter.exclude_masked(zdr_field)
        gatefilter.exclude_invalid(zdr_field)
    if (max_rhv is not None) and (rhv_field in radar.fields):
        gatefilter.exclude_above(rhv_field, max_rhv)
        gatefilter.exclude_masked(rhv_field)
        gatefilter.exclude_invalid(rhv_field)
    if refl_field in radar.fields:
        if min_refl is not None:
            gatefilter.exclude_below(refl_field, min_refl)
            gatefilter.exclude_masked(refl_field)
            gatefilter.exclude_invalid(refl_field)
        if max_refl is not None:
            gatefilter.exclude_above(refl_field, max_refl)
            gatefilter.exclude_masked(refl_field)
            gatefilter.exclude_invalid(refl_field)
    if (vel_lim is not None) and (vel_field in radar.fields):
        gatefilter.exclude_inside(vel_field, -vel_lim, vel_lim)
        gatefilter.exclude_masked(vel_field)
        gatefilter.exclude_invalid(vel_field)
    if rmin is not None or rmax is not None:
        mask = np.zeros((radar.nrays, radar.ngates), dtype='bool')
        if rmin is not None:
            ind = np.where(radar.range['data'] < rmin)[0]
            if ind.size > 0:
                mask[:, 0:ind[-1]+1] = True
        if rmax is not None:
            ind = np.where(radar.range['data'] > rmax)[0]
            if ind.size > 0:
                mask[:, ind[0]:] = True
        gatefilter.exclude_gates(mask, exclude_masked=True, op='or')
    if elmin is not None or elmax is not None:
        mask = np.zeros((radar.nrays, radar.ngates), dtype='bool')
        if elmin is not None:
            ind = np.where(radar.elevation['data'] < elmin)[0]
            if ind.size > 0:
                mask[0:ind[-1]+1, :] = True
        if elmax is not None:
            ind = np.where(radar.elevation['data'] > elmax)[0]
            if ind.size > 0:
                mask[ind[0]:, :] = True
        gatefilter.exclude_gates(mask, exclude_masked=True, op='or')

    return gatefilter


def snr_based_gate_filter(radar, snr_field=None, min_snr=10., max_snr=None):
    """
    Create a filter which removes undesired gates based on SNR.

    Parameters
    ----------
    radar : Radar
        Radar object from which the gate filter will be built.
    snr_field : str
        Name of the radar field which contains the signal to noise ratio.
        A value of None for will use the default field name as defined in
        the Py-ART configuration file.
    min_snr : float
        Minimum value for the SNR. Gates below this limits as well as gates
        which are masked or contain invalid values will be excluded and not
        used in calculation which use the filter. A value of None will disable
        filtering based upon the field including removing masked or
        gates with an invalid value. To disable the thresholding but retain
        the masked and invalid filter set the parameter to a value below the
        lowest value in the field.
    max_snr : float
        Maximum value for the SNR

    Returns
    -------
    gatefilter : :py:class:`GateFilter`
        A gate filter based upon the described criteria.  This can be
        used as a gatefilter parameter to various functions in pyart.correct.

    """
    # parse the field parameters
    if snr_field is None:
        snr_field = get_field_name('signal_to_noise_ratio')

    # make deepcopy of input radar (we do not want to modify the original)
    radar_aux = deepcopy(radar)

    # filter gates based upon field parameters
    gatefilter = GateFilter(radar_aux)

    if ((min_snr is not None or max_snr is not None) and
            (snr_field in radar_aux.fields)):
        gatefilter.exclude_masked(snr_field)
        gatefilter.exclude_invalid(snr_field)
        if min_snr is not None:
            gatefilter.exclude_below(snr_field, min_snr)
        if max_snr is not None:
            gatefilter.exclude_above(snr_field, max_snr)
    return gatefilter


def class_based_gate_filter(radar, field=None, kept_values=None):
    """
    Create a filter which removes undesired gates based on class values

    Parameters
    ----------
    radar : Radar
        Radar object from which the gate filter will be built.
    field : str
        Name of the radar field which contains the classification.
        A value of None for will use the default field name for the
        hydrometeor classification as defined in the Py-ART configuration
        file.
    kept_values : list of ints or none
        The class values to keep

    Returns
    -------
    gatefilter : :py:class:`GateFilter`
        A gate filter based upon the described criteria.  This can be
        used as a gatefilter parameter to various functions in pyart.correct.

    """
    # parse the field parameters
    if field is None:
        field = get_field_name('radar_echo_classification')

    # make deepcopy of input radar (we do not want to modify the original)
    radar_aux = deepcopy(radar)

    # filter gates based upon field parameters
    gatefilter = GateFilter(radar_aux)

    if (kept_values is not None) and (field in radar_aux.fields):
        gatefilter.exclude_masked(field)
        gatefilter.exclude_invalid(field)
        for value in kept_values:
            gatefilter.include_equal(field, value)

    return gatefilter


def visibility_based_gate_filter(radar, vis_field=None, min_vis=10.):
    """
    Create a filter which removes undesired gates based on visibility.

    Parameters
    ----------
    radar : Radar
        Radar object from which the gate filter will be built.
    vis_field : str
        Name of the radar field which contains the visibility.
        A value of None for will use the default field name as defined in
        the Py-ART configuration file.
    min_vis : float
        Minimum value for the visibility. Gates below this limits as well as
        gates which are masked or contain invalid values will be excluded and
        not used in calculation which use the filter. A value of None will
        disable filtering based upon the field including removing masked or
        gates with an invalid value. To disable the thresholding but retain
        the masked and invalid filter set the parameter to a value below the
        lowest value in the field.

    Returns
    -------
    gatefilter : :py:class:`GateFilter`
        A gate filter based upon the described criteria.  This can be
        used as a gatefilter parameter to various functions in pyart.correct.

    """
    # parse the field parameters
    if vis_field is None:
        vis_field = get_field_name('visibility')

    # make deepcopy of input radar (we do not want to modify the original)
    radar_aux = deepcopy(radar)

    # filter gates based upon field parameters
    gatefilter = GateFilter(radar_aux)

    if (min_vis is not None) and (vis_field in radar_aux.fields):
        gatefilter.exclude_below(vis_field, min_vis)
        gatefilter.exclude_masked(vis_field)
        gatefilter.exclude_invalid(vis_field)
    return gatefilter


def temp_based_gate_filter(radar, temp_field=None, min_temp=0.,
                           thickness=400., beamwidth=None):
    """
    Create a filter which removes undesired gates based on temperature. Used
    primarily to filter out the melting layer and gates above it.

    Parameters
    ----------
    radar : Radar
        Radar object from which the gate filter will be built.
    temp_field : str
        Name of the radar field which contains the temperature.
        A value of None for will use the default field name as defined in
        the Py-ART configuration file.
    min_temp : float
        Minimum value for the temperature in degrees. Gates below this limits
        as well as gates which are masked or contain invalid values will be
        excluded and not used in calculation which use the filter. A value of
        None will disable filtering based upon the field including removing
        masked or gates with an invalid value. To disable the thresholding but
        retain the masked and invalid filter set the parameter to a value
        below the lowest value in the field.
    thickness : float
        The estimated thickness of the melting layer in m.
    beamwidth : float
        The radar antenna 3 dB beamwidth [deg].

    Returns
    -------
    gatefilter : :py:class:`GateFilter`
        A gate filter based upon the described criteria. This can be
        used as a gatefilter parameter to various functions in pyart.correct.

    """
    # parse the field parameters
    if temp_field is None:
        temp_field = get_field_name('temperature')

    # make deepcopy of input radar (we do not want to modify the original)
    radar_aux = deepcopy(radar)

    # filter gates based upon field parameters
    gatefilter = GateFilter(radar_aux)

    if (min_temp is not None) and (temp_field in radar_aux.fields):
        gatefilter.exclude_below(temp_field, min_temp)
        gatefilter.exclude_masked(temp_field)
        gatefilter.exclude_invalid(temp_field)

    deltar = radar.range['data'][1]-radar.range['data'][0]
    if beamwidth is not None:
        beam_rad = beamwidth*np.pi/180.
    if thickness is not None:
        temp = radar_aux.fields[temp_field]
        temp['data'] = np.ma.masked_where(
            gatefilter.gate_excluded == 1, temp['data'])
        for ray in range(radar_aux.nrays):
            gate_h_ray = radar_aux.gate_altitude['data'][ray, :]
            # index of first excluded gate
            ind_r = np.where(gatefilter.gate_excluded[ray, :] == 1)[0]
            if ind_r.size > 0:
                # some gates are excluded: find the maximum height
                ind_r = ind_r[0]
                if beamwidth is None:
                    hmax = gate_h_ray[ind_r]-thickness
                else:
                    # consider also the radar volume
                    # maximum altitude at the end of the volume
                    if ind_r < radar_aux.ngates-2:
                        hmax = (
                            (gate_h_ray[ind_r] + gate_h_ray[ind_r+1])/2. -
                            thickness)
                    else:
                        hmax = gate_h_ray[ind_r]-thickness
                    beam_radius = (
                        (radar.range['data'][ind_r]+deltar/2.)*beam_rad/2.)
                    delta_h = (
                        beam_radius
                        * np.cos(radar.elevation['data'][ray]*np.pi/180.))
                    hmax -= delta_h

                ind_hmax = np.where(
                    radar_aux.gate_altitude['data'][ray, :] > hmax)[0]
                if ind_hmax.size > 0:
                    ind_hmax = ind_hmax[0]
                    temp['data'][ray, ind_hmax:] = np.ma.masked
        radar_aux.add_field(temp_field, temp, replace_existing=True)
        gatefilter = GateFilter(radar_aux)
        gatefilter.exclude_masked(temp_field)

    return gatefilter


def iso0_based_gate_filter(radar, iso0_field=None, max_h_iso0=0.,
                           thickness=400., beamwidth=None):
    """
    Create a filter which removes undesired gates based height over the iso0.
    Used primarily to filter out the melting layer and gates above it.

    Parameters
    ----------
    radar : Radar
        Radar object from which the gate filter will be built.
    iso0_field : str
        Name of the radar field which contains the height relative to the
        iso0. A value of None for will use the default field name as defined
        in the Py-ART configuration file.
    max_h_iso0 : float
        Maximum height relative to the iso0 in m. Gates below this limits
        as well as gates which are masked or contain invalid values will be
        excluded and not used in calculation which use the filter. A value of
        None will disable filtering based upon the field including removing
        masked or gates with an invalid value. To disable the thresholding but
        retain the masked and invalid filter set the parameter to a value
        below the lowest value in the field.
    thickness : float
        The estimated thickness of the melting layer in m.
    beamwidth : float
        The radar antenna 3 dB beamwidth [deg].

    Returns
    -------
    gatefilter : :py:class:`GateFilter`
        A gate filter based upon the described criteria. This can be
        used as a gatefilter parameter to various functions in pyart.correct.

    """
    # parse the field parameters
    if iso0_field is None:
        iso0_field = get_field_name('height_over_iso0')

    # make deepcopy of input radar (we do not want to modify the original)
    radar_aux = deepcopy(radar)

    # filter gates based upon field parameters
    gatefilter = GateFilter(radar_aux)

    if (max_h_iso0 is not None) and (iso0_field in radar_aux.fields):
        gatefilter.exclude_above(iso0_field, max_h_iso0)
        gatefilter.exclude_masked(iso0_field)
        gatefilter.exclude_invalid(iso0_field)

    deltar = radar.range['data'][1]-radar.range['data'][0]
    if beamwidth is not None:
        beam_rad = beamwidth*np.pi/180.
    if thickness is not None:
        iso0 = radar_aux.fields[iso0_field]
        iso0['data'] = np.ma.masked_where(
            gatefilter.gate_excluded == 1, iso0['data'])
        for ray in range(radar_aux.nrays):
            gate_h_ray = radar_aux.gate_altitude['data'][ray, :]
            # index of first excluded gate
            ind_r = np.where(gatefilter.gate_excluded[ray, :] == 1)[0]
            if ind_r.size > 0:
                # some gates are excluded: find the maximum height
                ind_r = ind_r[0]
                if beamwidth is None:
                    hmax = gate_h_ray[ind_r]-thickness
                else:
                    # consider also the radar volume
                    # maximum altitude at the end of the volume
                    if ind_r < radar_aux.ngates-2:
                        hmax = (
                            (gate_h_ray[ind_r] + gate_h_ray[ind_r+1])/2. -
                            thickness)
                    else:
                        hmax = gate_h_ray[ind_r]-thickness
                    beam_radius = (
                        (radar.range['data'][ind_r]+deltar/2.)*beam_rad/2.)
                    delta_h = (
                        beam_radius
                        * np.cos(radar.elevation['data'][ray]*np.pi/180.))
                    hmax -= delta_h

                ind_hmax = np.where(
                    radar_aux.gate_altitude['data'][ray, :] > hmax)[0]
                if ind_hmax.size > 0:
                    ind_hmax = ind_hmax[0]
                    iso0['data'][ray, ind_hmax:] = np.ma.masked
        radar_aux.add_field(iso0_field, iso0, replace_existing=True)
        gatefilter = GateFilter(radar_aux)
        gatefilter.exclude_masked(iso0_field)

    return gatefilter


class GateFilter(object):
    """
    A class for building a boolean arrays for filtering gates based on
    a set of condition typically based on the values in the radar fields.
    These filter can be used in various algorithms and calculations within
    Py-ART.

    See :py:func:`pyart.correct.GateFilter.exclude_below` for method
    parameter details.

    Parameters
    ----------
    radar : Radar
        Radar object from which gate filter will be build.
    exclude_based : bool, optional
        True, the default and suggested method, will begin with all gates
        included and then use the exclude methods to exclude gates based on
        conditions. False will begin with all gates excluded from which
        a set of gates to include should be set using the include methods.

    Attributes
    ----------
    gate_excluded : array, dtype=bool
        Boolean array indicating if a gate should be excluded from a
        calculation. Elements marked True indicate the corresponding gate
        should be excluded. Those marked False should be included.
        This is read-only attribute, any changes to the array will NOT
        be reflected in gate_included and will be lost when the attribute is
        accessed again.
    gate_included : array, dtype=bool
        Boolean array indicating if a gate should be included in a
        calculation. Elements marked True indicate the corresponding gate
        should be include. Those marked False should be excluded.
        This is read-only attribute, any changes to the array will NOT
        be reflected in gate_excluded and will be lost when the attribute is
        accessed again.

    Examples
    --------
    >>> import pyart
    >>> radar = pyart.io.read('radar_file.nc')
    >>> gatefilter = pyart.correct.GateFilter(radar)
    >>> gatefilter.exclude_below('reflectivity', 10)
    >>> gatefilter.exclude_below('normalized_coherent_power', 0.75)

    """

    def __init__(self, radar, exclude_based=True):
        """ initialize """
        self._radar = radar
        shape = (radar.nrays, radar.ngates)
        if exclude_based:
            # start with all gates included, exclude gates based on a set
            # of rules using the exclude_ methods.
            self._gate_excluded = np.zeros(shape, dtype=np.bool)
        else:
            # start with all gates excluded, include gates based on a set
            # of rules using the include_ methods.
            self._gate_excluded = np.ones(shape, dtype=np.bool)

    # Implemetation is based on marking excluded gates stored in the private
    # _gate_excluded attribute. The gate_included attribute can be found
    # by taking the ones complement of gates_included.

    def copy(self):
        """ Return a copy of the gatefilter. """
        a = GateFilter(self._radar)
        a._gate_excluded = self._gate_excluded.copy()
        return a

    @property
    def gate_included(self):
        """ Return a copy of the included gates. """
        return ~self._gate_excluded.copy()

    @property
    def gate_excluded(self):
        """ Return a copy of the excluded gates. """
        return self._gate_excluded.copy()

    def _get_fdata(self, field):
        """ Check that the field exists and retrieve field data. """
        self._radar.check_field_exists(field)
        return self._radar.fields[field]['data']

    def _merge(self, marked, op, exclude_masked):
        """ Merge an array of marked gates with the exclude array. """
        # exclude masked elements in marked by replacing them with the value
        # of the exclude_masked flag.  This does nothing if marked is a
        # non-masked array.
        if exclude_masked not in [True, False]:
            raise ValueError("exclude_masked must be 'True' or 'False'")
        marked = np.ma.filled(marked, exclude_masked)

        # merge array of marked gates with existing excluded gates
        # using the specified operation.
        if op == 'or':
            self._gate_excluded = np.logical_or(self._gate_excluded, marked)
        elif op == 'and':
            self._gate_excluded = np.logical_and(self._gate_excluded, marked)
        elif op == 'new':
            self._gate_excluded = marked
        else:
            raise ValueError("invalid 'op' parameter: ", op)

    ###################
    # exclude methods #
    ###################

    def exclude_transition(self, trans_value=1, exclude_masked=True, op='or'):
        """
        Exclude all gates in rays marked as in transition between sweeps.

        Exclude all gates in rays marked as "in transition" by the
        antenna_transition attribute of the radar used to construct the filter.
        If no antenna transition information is available no gates are
        excluded.

        Parameters
        ----------
        trans_value : int, optional
            Value used in the antenna transition data to indicate that the
            instrument was between sweeps (in transition) during the collection
            of a specific ray. Typically a value of 1 is used to indicate this
            transition and the default can be used in these cases.
        exclude_masked : bool, optional
            True to filter masked values in antenna_transition if the data is
            a masked array, False to include any masked values.
        op : {'and', 'or', 'new'}
            Operation to perform when merging the existing set of excluded
            gates with the excluded gates from the current operation.
            'and' will perform a logical AND operation, 'or' a logical OR,
            and 'new' will replace the existing excluded gates with the one
            generated here. 'or', the default for exclude methods, is
            typically desired when building up a set of conditions for
            excluding gates where the desired effect is to exclude gates which
            meet any of the conditions. 'and', the default for include
            methods, is typically desired when building up a set of conditions
            where the desired effect is to include gates which meet any of the
            conditions.  Note that the 'and' method MAY results in including
            gates which have previously been excluded because they were masked
            or invalid.

        """
        marked = np.zeros_like(self._gate_excluded)
        if self._radar.antenna_transition is not None:
            transition_data = self._radar.antenna_transition['data']
            in_transition = transition_data == trans_value
            marked[in_transition] = True
        return self._merge(marked, op, exclude_masked)

    def exclude_below(self, field, value, exclude_masked=True, op='or',
                      inclusive=False):
        """
        Exclude gates where a given field is below a given value.

        Parameters
        ----------
        field : str
            Name of field compared against the value.
        value : float
            Gates with a value below this value in the specified field will
            be marked for exclusion in the filter.
        exclude_masked : bool, optional
            True to filter masked values in the specified field if the data is
            a masked array, False to include any masked values.
        op : {'and', 'or', 'new'}
            Operation to perform when merging the existing set of excluded
            gates with the excluded gates from the current operation.
            'and' will perform a logical AND operation, 'or' a logical OR,
            and 'new' will replace the existing excluded gates with the one
            generated here. 'or', the default for exclude methods, is
            typically desired when building up a set of conditions for
            excluding gates where the desired effect is to exclude gates which
            meet any of the conditions. 'and', the default for include
            methods, is typically desired when building up a set of conditions
            where the desired effect is to include gates which meet any of the
            conditions.  Note that the 'and' method MAY results in including
            gates which have previously been excluded because they were masked
            or invalid.
        inclusive : bool
            Indicates whether the specified value should also be excluded.

        """
        if inclusive:
            marked = self._get_fdata(field) <= value
        else:
            marked = self._get_fdata(field) < value
        return self._merge(marked, op, exclude_masked)

    def exclude_above(self, field, value, exclude_masked=True, op='or',
                      inclusive=False):
        """ Exclude gates where a given field is above a given value. """
        if inclusive:
            marked = self._get_fdata(field) >= value
        else:
            marked = self._get_fdata(field) > value
        return self._merge(marked, op, exclude_masked)

    def exclude_inside(self, field, v1, v2, exclude_masked=True, op='or',
                       inclusive=True):
        """ Exclude gates where a given field is inside a given interval. """
        if v2 < v1:
            (v1, v2) = (v2, v1)
        fdata = self._get_fdata(field)
        if inclusive:
            marked = (fdata >= v1) & (fdata <= v2)
        else:
            marked = (fdata > v1) & (fdata < v2)
        return self._merge(marked, op, exclude_masked)

    def exclude_outside(self, field, v1, v2, exclude_masked=True, op='or',
                        inclusive=False):
        """ Exclude gates where a given field is outside a given interval. """
        if v2 < v1:
            (v1, v2) = (v2, v1)
        fdata = self._get_fdata(field)
        if inclusive:
            marked = (fdata <= v1) | (fdata >= v2)
        else:
            marked = (fdata < v1) | (fdata > v2)
        return self._merge(marked, op, exclude_masked)

    def exclude_equal(self, field, value, exclude_masked=True, op='or'):
        """ Exclude gates where a given field is equal to a value. """
        marked = (self._get_fdata(field) == value)
        return self._merge(marked, op, exclude_masked)

    def exclude_not_equal(self, field, value, exclude_masked=True, op='or'):
        """ Exclude gates where a given field is not equal to a value. """
        marked = (self._get_fdata(field) != value)
        return self._merge(marked, op, exclude_masked)

    def exclude_all(self):
        """ Exclude all gates. """
        self._gate_excluded = np.ones_like(self._gate_excluded)

    def exclude_none(self):
        """ Exclude no gates, include all gates. """
        self._gate_excluded = np.zeros_like(self._gate_excluded)

    def exclude_masked(self, field, exclude_masked=True, op='or'):
        """ Exclude gates where a given field is masked. """
        marked = np.ma.getmaskarray(self._get_fdata(field))
        return self._merge(marked, op, exclude_masked)

    def exclude_invalid(self, field, exclude_masked=True, op='or'):
        """
        Exclude gates where an invalid value occurs in a field (NaNs or infs).
        """
        marked = ~np.isfinite(self._get_fdata(field))
        return self._merge(marked, op, exclude_masked)

    def exclude_gates(self, mask, exclude_masked=True, op='or'):
        """
        Exclude gates where a given mask is equal True.

        Parameters
        ----------
        mask : numpy array
            Boolean numpy array with same shape as a field array.
        exclude_masked : bool, optional
            True to filter masked values in the specified mask if it is
            a masked array, False to include any masked values.
        op : {'and', 'or', 'new'}
            Operation to perform when merging the existing set of excluded
            gates with the excluded gates from the current operation.
            'and' will perform a logical AND operation, 'or' a logical OR,
            and 'new' will replace the existing excluded gates with the one
            generated here. 'or', the default for exclude methods, is
            typically desired when building up a set of conditions for
            excluding gates where the desired effect is to exclude gates which
            meet any of the conditions. 'and', the default for include
            methods, is typically desired when building up a set of conditions
            where the desired effect is to include gates which meet any of the
            conditions.  Note that the 'and' method MAY results in including
            gates which have previously been excluded because they were masked
            or invalid.

        """
        fdata = next(iter(self._radar.fields.values()))['data']
        if mask.shape != fdata.shape:
            raise ValueError("mask array must be the same size as a field.")
        marked = np.array(mask, dtype='bool')
        return self._merge(marked, op, exclude_masked)

    ####################
    # include_ methods #
    ####################

    def include_not_transition(
            self, trans_value=0, exclude_masked=True, op='and'):
        """
        Include all gates in rays not marked as in transition between sweeps.

        Include all gates in rays not marked as "in transition" by the
        antenna_transition attribute of the radar used to construct the filter.
        If no antenna transition information is available all gates are
        included.

        Parameters
        ----------
        trans_value : int, optional
            Value used in the antenna transition data to indicate that the
            instrument is not between sweeps (in transition) during the
            collection of a specific ray. Typically a value of 0 is used to
            indicate no transition and the default can be used in these cases.
        exclude_masked : bool, optional
            True to filter masked values in antenna_transition if the data is
            a masked array, False to include any masked values.
        op : {'and', 'or', 'new'}
            Operation to perform when merging the existing set of excluded
            gates with the excluded gates from the current operation.
            'and' will perform a logical AND operation, 'or' a logical OR,
            and 'new' will replace the existing excluded gates with the one
            generated here. 'or', the default for exclude methods, is
            typically desired when building up a set of conditions for
            excluding gates where the desired effect is to exclude gates which
            meet any of the conditions. 'and', the default for include
            methods, is typically desired when building up a set of conditions
            where the desired effect is to include gates which meet any of the
            conditions.  Note that the 'or' method MAY results in excluding
            gates which have previously been included.

        """
        if self._radar.antenna_transition is None:
            include = np.ones_like(self._gate_excluded)  # include all gates
        else:
            include = np.zeros_like(self._gate_excluded)
            transition_data = self._radar.antenna_transition['data']
            not_in_transition = (transition_data == trans_value)
            include[not_in_transition] = True
        return self._merge(~include, op, exclude_masked)

    def include_below(self, field, value, exclude_masked=True, op='and',
                      inclusive=False):
        """ Include gates where a given field is below a given value. """
        if inclusive:
            marked = self._get_fdata(field) <= value
        else:
            marked = self._get_fdata(field) < value
        self._merge(~marked, op, exclude_masked)

    def include_above(self, field, value, exclude_masked=True, op='and',
                      inclusive=False):
        """ Include gates where a given field is above a given value. """
        if inclusive:
            marked = self._get_fdata(field) >= value
        else:
            marked = self._get_fdata(field) > value
        self._merge(~marked, op, exclude_masked)

    def include_inside(self, field, v1, v2, exclude_masked=True, op='and',
                       inclusive=True):
        """ Include gates where a given field is inside a given interval. """
        if v2 < v1:
            (v1, v2) = (v2, v1)
        fdata = self._get_fdata(field)
        if inclusive:
            marked = (fdata >= v1) & (fdata <= v2)
        else:
            marked = (fdata > v1) & (fdata < v2)
        return self._merge(~marked, op, exclude_masked)

    def include_outside(self, field, v1, v2, exclude_masked=True, op='and',
                        inclusive=False):
        """ Include gates where a given field is outside a given interval. """
        if v2 < v1:
            (v1, v2) = (v2, v1)
        fdata = self._get_fdata(field)
        if inclusive:
            marked = (fdata <= v1) | (fdata >= v2)
        else:
            marked = (fdata < v1) | (fdata > v2)
        return self._merge(~marked, op, exclude_masked)

    def include_equal(self, field, value, exclude_masked=True, op='and'):
        """ Include gates where a given field is equal to a value. """
        marked = (self._get_fdata(field) == value)
        return self._merge(~marked, op, exclude_masked)

    def include_not_equal(self, field, value, exclude_masked=True, op='and'):
        """ Include gates where a given field is not equal to a value. """
        marked = (self._get_fdata(field) != value)
        return self._merge(~marked, op, exclude_masked)

    def include_all(self):
        """ Include all gates. """
        self._gate_excluded = np.zeros_like(self._gate_excluded)

    def include_none(self):
        """ Include no gates, exclude all gates. """
        self._gate_excluded = np.ones_like(self._gate_excluded)

    def include_not_masked(self, field, exclude_masked=True, op='and'):
        """ Include gates where a given field in not masked. """
        marked = np.ma.getmaskarray(self._get_fdata(field))
        return self._merge(marked, op, exclude_masked)

    def include_valid(self, field, exclude_masked=True, op='and'):
        """
        Include gates where a valid value occurs in a field (not NaN or inf).
        """
        marked = np.isfinite(self._get_fdata(field))
        return self._merge(~marked, op, exclude_masked)

    def include_gates(self, mask, exclude_masked=True, op='and'):
        """
        Include gates where a given mask is equal True.

        Parameters
        ----------
        mask : numpy array
            Boolean numpy array with same shape as a field array.
        exclude_masked : bool, optional
            True to filter masked values in the specified mask if it is
            a masked array, False to include any masked values.
        op : {'and', 'or', 'new'}
            Operation to perform when merging the existing set of excluded
            gates with the excluded gates from the current operation.
            'and' will perform a logical AND operation, 'or' a logical OR,
            and 'new' will replace the existing excluded gates with the one
            generated here. 'or', the default for exclude methods, is
            typically desired when building up a set of conditions for
            excluding gates where the desired effect is to exclude gates which
            meet any of the conditions. 'and', the default for include
            methods, is typically desired when building up a set of conditions
            where the desired effect is to include gates which meet any of the
            conditions.  Note that the 'or' method MAY results in excluding
            gates which have previously been included.

        """
        fdata = next(iter(self._radar.fields.values()))['data']
        if mask.shape != fdata.shape:
            raise ValueError("Mask array must be the same size as a field.")
        marked = ~np.array(mask, dtype='bool')
        return self._merge(marked, op, exclude_masked)



"""
pyart.io.grid_io
================

Reading and writing Grid objects.

.. autosummary::
    :toctree: generated/

    read_grid
    write_grid
    _make_coordinatesystem_dict

"""

import datetime
import warnings

import netCDF4
import numpy as np

from ..core.grid import Grid
from .cfradial import _ncvar_to_dict, _create_ncvar
from .common import _test_arguments


def read_grid(filename, exclude_fields=None, include_fields=None, **kwargs):
    """
    Read a netCDF grid file produced by Py-ART.

    Parameters
    ----------
    filename : str
        Filename of netCDF grid file to read. This file must have been
        produced by :py:func:`write_grid` or have identical layout.

    Other Parameters
    ----------------
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.

    Returns
    -------
    grid : Grid
        Grid object containing gridded data.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    if exclude_fields is None:
        exclude_fields = []

    reserved_variables = [
        'time', 'x', 'y', 'z',
        'origin_latitude', 'origin_longitude', 'origin_altitude',
        'point_x', 'point_y', 'point_z', 'projection',
        'point_latitude', 'point_longitude', 'point_altitude',
        'radar_latitude', 'radar_longitude', 'radar_altitude',
        'radar_name', 'radar_time', 'base_time', 'time_offset',
        'ProjectionCoordinateSystem']

    dset = netCDF4.Dataset(filename, mode='r')

    # metadata
    metadata = dict([(k, getattr(dset, k)) for k in dset.ncattrs()])

    # required reserved variables
    time = _ncvar_to_dict(dset.variables['time'])
    origin_latitude = _ncvar_to_dict(dset.variables['origin_latitude'])
    origin_longitude = _ncvar_to_dict(dset.variables['origin_longitude'])
    origin_altitude = _ncvar_to_dict(dset.variables['origin_altitude'])
    x = _ncvar_to_dict(dset.variables['x'])
    y = _ncvar_to_dict(dset.variables['y'])
    z = _ncvar_to_dict(dset.variables['z'])

    # projection
    projection = _ncvar_to_dict(dset.variables['projection'])
    projection.pop('data')
    # map _include_lon_0_lat_0 key to bool type
    if '_include_lon_0_lat_0' in projection:
        v = projection['_include_lon_0_lat_0']
        projection['_include_lon_0_lat_0'] = {'true': True, 'false': False}[v]

    # read in the fields
    fields = {}

    # fields in the file has a shape of (1, nz, ny, nx) with the leading 1
    # indicating time but should shaped (nz, ny, nx) in the Grid object
    field_shape = tuple([len(dset.dimensions[d]) for d in ['z', 'y', 'x']])
    field_shape_with_time = (1, ) + field_shape

    # check all non-reserved variables, those with the correct shape
    # are added to the field dictionary, if a wrong sized field is
    # detected a warning is raised
    field_keys = [k for k in dset.variables if k not in reserved_variables]
    for field in field_keys:
        if field in exclude_fields:
            continue
        if include_fields is not None:
            if field not in include_fields:
                continue
        field_dic = _ncvar_to_dict(dset.variables[field])
        if field_dic['data'].shape == field_shape_with_time:
            field_dic['data'].shape = field_shape
            fields[field] = field_dic
        else:
            bad_shape = field_dic['data'].shape
            warnings.warn(
                'Field %s skipped due to incorrect shape %s'
                % (field, bad_shape))

    # radar_ variables
    if 'radar_latitude' in dset.variables:
        radar_latitude = _ncvar_to_dict(dset.variables['radar_latitude'])
    else:
        radar_latitude = None

    if 'radar_longitude' in dset.variables:
        radar_longitude = _ncvar_to_dict(dset.variables['radar_longitude'])
    else:
        radar_longitude = None

    if 'radar_altitude' in dset.variables:
        radar_altitude = _ncvar_to_dict(dset.variables['radar_altitude'])
    else:
        radar_altitude = None

    if 'radar_name' in dset.variables:
        radar_name = _ncvar_to_dict(dset.variables['radar_name'])
    else:
        radar_name = None

    if 'radar_time' in dset.variables:
        radar_time = _ncvar_to_dict(dset.variables['radar_time'])
    else:
        radar_time = None

    dset.close()

    return Grid(
        time, fields, metadata,
        origin_latitude, origin_longitude, origin_altitude, x, y, z,
        projection=projection,
        radar_latitude=radar_latitude, radar_longitude=radar_longitude,
        radar_altitude=radar_altitude, radar_name=radar_name,
        radar_time=radar_time)


def write_grid(filename, grid, format='NETCDF4',
               write_proj_coord_sys=True, proj_coord_sys=None,
               arm_time_variables=False, arm_alt_lat_lon_variables=False,
               write_point_x_y_z=False, write_point_lon_lat_alt=False):
    """
    Write a Grid object to a CF-1.5 and ARM standard netCDF file.

    To control how the netCDF variables are created, set any of the following
    keys in the grid attribute dictionaries.

        * _Zlib
        * _DeflateLevel
        * _Shuffle
        * _Fletcher32
        * _Continguous
        * _ChunkSizes
        * _Endianness
        * _Least_significant_digit
        * _FillValue

    See the netCDF4 documentation for details on these settings.

    Parameters
    ----------
    filename : str
        Filename to save grid to.
    grid : Grid
        Grid object to write.
    format : str, optional
        netCDF format, one of 'NETCDF4', 'NETCDF4_CLASSIC',
        'NETCDF3_CLASSIC' or 'NETCDF3_64BIT'. See netCDF4 documentation for
        details.
    write_proj_coord_sys bool, optional
        True to write information on the coordinate transform used in the map
        projection to the ProjectionCoordinateSystem variable following the CDM
        Object Model. The resulting file should be interpreted as containing
        geographic grids by tools which use the Java NetCDF library
        (THREDDS, toolsUI, etc).
    proj_coord_sys : dict or None, optional
        Dictionary of parameters which will be written to the
        ProjectionCoordinateSystem NetCDF variable if write_proj_coord_sys is
        True. A value of None will attempt to generate an appropriate
        dictionary by examining the projection attribute of the grid object.
        If the projection is not understood a warnings will be issued.
    arm_time_variables : bool, optional
        True to write the ARM standard time variables base_time and
        time_offset. False will not write these variables.
    arm_alt_lat_lon_variables : bool, optional
        True to write the ARM standard alt, lat, lon variables.
        False will not write these variables.
    write_point_x_y_z : bool, optional
        True to include the point_x, point_y and point_z variables in the
        written file, False will not write these variables.
    write_point_lon_lat_alt : bool, optional
        True to include the point_longitude, point_latitude and point_altitude
        variables in the written file, False will not write these variables.

    """
    dset = netCDF4.Dataset(filename, mode='w', format=format)

    # create dimensions
    dset.createDimension('time', None)
    dset.createDimension('z', grid.nz)
    dset.createDimension('y', grid.ny)
    dset.createDimension('x', grid.nx)
    if grid.nradar != 0:
        dset.createDimension('nradar', grid.nradar)
        if grid.radar_name is not None:
            # a length of at least 1 is required for the dimension
            lenlist = []
            for rname in grid.radar_name['data']:
                lenlist.append(len(rname))
            lenlist.append(1)
            nradar_str_length = np.max(lenlist)
            # nradar_str_length = max(len(grid.radar_name['data'][0]), 1)
            dset.createDimension('nradar_str_length', nradar_str_length)

    # required variables
    _create_ncvar(grid.time, dset, 'time', ('time', ))
    _create_ncvar(grid.x, dset, 'x', ('x', ))
    _create_ncvar(grid.y, dset, 'y', ('y', ))
    _create_ncvar(grid.z, dset, 'z', ('z', ))
    _create_ncvar(grid.origin_latitude, dset, 'origin_latitude', ('time', ))
    _create_ncvar(grid.origin_longitude, dset, 'origin_longitude', ('time', ))
    _create_ncvar(grid.origin_altitude, dset, 'origin_altitude', ('time', ))

    # write the projection dictionary as a scalar
    projection = grid.projection.copy()
    projection['data'] = np.array(1, dtype='int32')
    # NetCDF does not support boolean attribute, covert to string
    if '_include_lon_0_lat_0' in projection:
        include = projection['_include_lon_0_lat_0']
        projection['_include_lon_0_lat_0'] = ['false', 'true'][include]
    _create_ncvar(projection, dset, 'projection', ())

    # set the default projection coordinate system if requested
    if write_proj_coord_sys:
        if proj_coord_sys is None:
            # determine coordinate system automatically
            proj_coord_sys = _make_coordinatesystem_dict(grid)

        if proj_coord_sys is None:
            warnings.warn(
                'Cannot determine ProjectionCoordinateSystem parameter for ' +
                'the given projection, the file will not be written ' +
                'without this information.')
        else:
            proj_coord_sys['data'] = np.array(1, dtype='int32')
            _create_ncvar(
                proj_coord_sys, dset, 'ProjectionCoordinateSystem', ())

    # radar_ attributes
    radar_attr_names = [
        'radar_latitude', 'radar_longitude', 'radar_altitude', 'radar_time']
    for attr_name in radar_attr_names:
        attr = getattr(grid, attr_name)
        if attr is not None:
            _create_ncvar(attr, dset, attr_name, ('nradar', ))

    if grid.radar_name is not None:
        _create_ncvar(grid.radar_name, dset, 'radar_name',
                      ('nradar', 'nradar_str_length'))

    # create ARM time variables base_time and time_offset, if requested
    if arm_time_variables:
        time = grid.time
        dt = netCDF4.num2date(time['data'][0], time['units'])
        td = dt - datetime.datetime.utcfromtimestamp(0)
        td = td.seconds + td.days * 24 * 3600

        base_time = {
            'data': np.array([td], dtype=np.int32),
            'string': dt.strftime('%d-%b-%Y,%H:%M:%S GMT'),
            'units': 'seconds since 1970-1-1 0:00:00 0:00',
            'ancillary_variables': 'time_offset',
            'long_name': 'Base time in Epoch',
        }
        _create_ncvar(base_time, dset, 'base_time', ())

        time_offset = {
            'data': np.array(time['data'], dtype=np.float64),
            'long_name': 'Time offset from base_time',
            'units': time['units'].replace('T', ' ').replace('Z', ''),
            'ancillary_variables': 'time_offset',
            'calendar': 'gregorian',
        }
        _create_ncvar(time_offset, dset, 'time_offset', ('time', ))

    # create ARM alt, lat, lon variables, if requested
    if arm_alt_lat_lon_variables:
        alt = {
            'data': np.array([grid.origin_altitude['data']], dtype=np.float64),
            'standard_name': 'Altitude',
            'units': 'm',
            'long_name': 'Altitude above mean sea level',
        }
        _create_ncvar(alt, dset, 'alt', ())

        lat = {
            'data': np.array([grid.origin_latitude['data']], dtype=np.float64),
            'standard_name': 'Latitude',
            'units': 'degree_N',
            'long_name': 'North Latitude',
            'valid_min': -90.,
            'valid_max': 90.,
        }
        _create_ncvar(lat, dset, 'lat', ())

        lon = {
            'data': np.array([grid.origin_longitude['data']], dtype=np.float64),
            'standard_name': 'Longitude',
            'units': 'degree_E',
            'long_name': 'East Longitude',
            'valid_min': -180.,
            'valid_max': 180.,
        }
        _create_ncvar(lon, dset, 'lon', ())

    # optionally write point_ variables
    if write_point_x_y_z:
        _create_ncvar(grid.point_x, dset, 'point_x', ('z', 'y', 'x'))
        _create_ncvar(grid.point_y, dset, 'point_y', ('z', 'y', 'x'))
        _create_ncvar(grid.point_z, dset, 'point_z', ('z', 'y', 'x'))
    if write_point_lon_lat_alt:
        dims = ('z', 'y', 'x')
        _create_ncvar(grid.point_latitude, dset, 'point_latitude', dims)
        _create_ncvar(grid.point_longitude, dset, 'point_longitude', dims)
        _create_ncvar(grid.point_altitude, dset, 'point_altitude', dims)

    # field variables
    for field, field_dic in grid.fields.items():
        # append 1, to the shape of all data to indicate the time var.
        field_dic['data'].shape = (1, ) + field_dic['data'].shape
        _create_ncvar(field_dic, dset, field, ('time', 'z', 'y', 'x'))
        field_dic['data'].shape = field_dic['data'].shape[1:]

    # metadata
    for k, v in grid.metadata.items():
        setattr(dset, k, v)

    # Add Conventions if not already present
    if 'Conventions' not in dset.ncattrs():
        dset.setncattr('Conventions', 'PyART_GRID-1.1')

    dset.close()


def _make_coordinatesystem_dict(grid):
    """
    Return a dictionary containing parameters for a coordinate transform.

    Examine the grid projection attribute and other grid attributes to
    return a dictionary containing parameters which can be written to a netCDF
    variable to specify a horizontal coordinate transform recognized by
    Unidata's CDM. Return None when the projection defined in the grid
    cannot be mapped to a CDM coordinate transform.
    """
    projection = grid.projection
    origin_latitude = grid.origin_latitude['data'][0]
    origin_longitude = grid.origin_longitude['data'][0]
    cdm_transform = {
        'latitude_of_projection_origin': origin_latitude,
        'longitude_of_projection_origin': origin_longitude,
        '_CoordinateTransformType': 'Projection',
        '_CoordinateAxes': 'x y z time',
        '_CoordinateAxesTypes': 'GeoX GeoY Height Time',
    }

    if projection['proj'] == 'ortho':
        cdm_transform['grid_mapping_name'] = 'orthographic'

    elif projection['proj'] == 'laea':
        cdm_transform['grid_mapping_name'] = 'lambert_azimuthal_equal_area'

    elif projection['proj'] in ['aeqd', 'pyart_aeqd']:
        cdm_transform['grid_mapping_name'] = 'azimuthal_equidistant'
        # CDM uses a ellipsoid where as PyProj uses a sphere by default,
        # therefore there will be slight differences in these transforms
        cdm_transform['semi_major_axis'] = 6370997.0
        cdm_transform['inverse_flattening'] = 298.25  # proj uses a sphere
        cdm_transform['longitude_of_prime_meridian'] = 0.0
        cdm_transform['false_easting'] = 0.0
        cdm_transform['false_northing'] = 0.0

    elif projection['proj'] == 'tmerc':
        cdm_transform['grid_mapping_name'] = 'transverse_mercator'
        cdm_transform['longitude_of_central_meridian'] = origin_longitude
        cdm_transform['scale_factor_at_central_meridian'] = 1.00

    elif projection['proj'] == 'lcc':
        cdm_transform['grid_mapping_name'] = 'lambert_conformal_conic'
        cdm_transform['standard_parallel'] = origin_latitude
        cdm_transform['longitude_of_central_meridian'] = origin_longitude

    elif projection['proj'] == 'aea':
        cdm_transform['grid_mapping_name'] = 'albers_conical_equal_area'
        cdm_transform['standard_parallel'] = origin_latitude
        cdm_transform['longitude_of_central_meridian'] = origin_longitude

    elif projection['proj'] == 'stere':
        cdm_transform['grid_mapping_name'] = 'stereographic'
        cdm_transform['scale_factor_at_projection_origin'] = 1.00

    elif projection['proj'] in ['npstere', 'spstere']:
        cdm_transform['grid_mapping_name'] = 'polar_stereographic'
        cdm_transform['standard_parallel'] = origin_latitude

    # 'cea' may be able to map to 'lambert_cylindrical_equal_area' and
    # 'merc' to 'mercator' but both projections seems to always be
    # centered at the equator regardless of the value of the
    # standard_parallel parameter
    else:
        cdm_transform = None

    return cdm_transform



# -*- coding: utf-8 -*-
"""
metranet python library
================

Information about the variable names, types and length that are stored in
the metranet files

History

      V0.1 20190306 wod first prototype

"""


#List of abbreviations
# i = signed integer
# I = unsigned integer
# B = byte
# c = character
# d = double
# f = float (single)
# s = string
# H = unsigned short
# h = short
# L = unsigned long
# Q = unsigned long long


# Here are the byte sizes for all types
BYTE_SIZES = {'i':4, 'f':4, 'B':1, 'I':4, 'H':2, 'd': 8, 's': 1, 'L':4, 'c':1,
              'b':1, 'h':2, 'Q':8, 'p': 1}


'''
===============================================================================
M-files
===============================================================================
'''


###############################################################################
# SWEEP_HEADER INFO

MSWEEP_HEADER = {}

MSWEEP_HEADER['names'] = [
    'field',
    'version',
    'spare',
    'length',
    'radarname',
    'scanname',
    'radarlat',
    'radarlon',
    'radarheight',
    'sequencesweep',
    'currentsweep',
    'totalsweep',
    'antmode',
    'priority',
    'quality',
    'spare',
    'repeattime',
    'nummoments',
    'gatewidth',
    'frequency',
    'pulsewidth',
    'startrange',
    'metadatasize',
    'moments',
    'metadata']

# Sweep header entries variable types (see https://docs.python.org/3/library/struct.html)
MSWEEP_HEADER['type'] = [
    's',
    'b',
    'b',
    'L',
    's',
    's',
    'f',
    'f',
    'f',
    'b',
    'b',
    'b',
    'b',
    'b',
    'b',
    'b',
    'H',
    'H',
    'f',
    'f',
    'f',
    'f',
    'L',
    'b',
    's']

'''
Note: length of sweep header entries in multiples of type byte sizes. ex if
type = 'd' and len = 3, the total size in bytes will be
8 (bytes/double) * 3 = 18 bytes. When the length is a list ex.
[4, 'variable'], it means that the size of this field will be 4 x the value of
the 'variable' key in the SWEEP_HEADER x the size in bytes of the type.
'''

MSWEEP_HEADER['len'] = [
    4,
    1,
    3,
    1,
    16,
    16,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    2,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    [48, 'nummoments'],
    [1, 'metadatasize']]


##############################################################################
### MOMENT_INFO STRUCTURE INFO

MMOMENT_INFO_STRUCTURE = {}

MMOMENT_INFO_STRUCTURE['names'] = [
    'moment_uuid',
    'data_format',
    'num_bytes',
    'spare1',
    'name',
    'unit',
    'a',
    'b',
    'c',
    'scale_type',
    'spare2']

MMOMENT_INFO_STRUCTURE['type'] = [
    'L',
    'b',
    'b',
    'b',
    's',
    's',
    'f',
    'f',
    'f',
    'b',
    'b']

MMOMENT_INFO_STRUCTURE['len'] = [
    1,
    1,
    1,
    2,
    12,
    12,
    1,
    1,
    1,
    1,
    3]

##############################################################################
### RAY HEADER INFO

MRAY_HEADER = {}

MRAY_HEADER['names'] = [
    'length',
    'startangle_az',
    'startangle_el',
    'endangle_az',
    'endangle_el',
    'sequence',
    'numpulses',
    'databytes',
    'prf',
    'datatime',
    'dataflags',
    'metadatasize',
    'metadata']

MRAY_HEADER['type'] = [
    'I',
    'H',
    'h',
    'H',
    'h',
    'H',
    'H',
    'I',
    'f',
    'Q',
    'I',
    'I',
    's']

MRAY_HEADER['len'] = [
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    [1, 'metadatasize']]

##############################################################################
### DATA MOMENT HEADER INFO

MMOMENT_HEADER = {}

MMOMENT_HEADER['names'] = [
    'momentuuid',
    'datasize']

MMOMENT_HEADER['type'] = [
    'L',
    'I']

MMOMENT_HEADER['len'] = [
    1,
    1]

'''
==============================================================================
P-files
==============================================================================
'''

##############################################################################
### RAY HEADER INFO
PRAY_HEADER = {}

PRAY_HEADER = {}

PRAY_HEADER['names'] = [
    'dataformat',
    'bitmask',
    'scanid',
    'hostid',
    'startangle_el',
    'startangle_az',
    'endangle_el',
    'endangle_az',
    'antmode',
    'totalsweep',
    'currentsweep',
    'endofsweep',
    'sequence',
    'totalrecord',
    'pulses',
    'numgates',
    'databytes',
    'dataflag',
    'datatime_residue',
    'datatime',
    'repeattime',
    'compressed',
    'priority',
    'nyquest',
    'gatewidth',
    'wnyquest',
    'startrange']

PRAY_HEADER['type'] = [
    'b',
    'b',
    's',
    'I',
    'h',
    'H',
    'h',
    'H',
    'b',
    'b',
    'b',
    'b',
    'h',
    'h',
    'h',
    'h',
    'i',
    'H',
    'h',
    'I',
    'h',
    'b',
    'b',
    'f',
    'f',
    'f',
    'f']

PRAY_HEADER['len'] = [
    1,
    3,
    4,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1]

##############################################################################
### DATA MOMENT SIZES
PMOMENTS = {}
PMOMENTS['names'] = {}
PMOMENTS['names']['M'] = [
    'VEL', 'ZH', 'ZV', 'ZDR', 'RHO', 'PHI', 'ST1', 'ST2', 'WBN', 'MPH']
PMOMENTS['names']['L'] = [
    'WID', 'VEL', 'ZH', 'ZV', 'ZDR', 'RHO', 'PHI', 'ZHC', 'ZVC']
PMOMENTS['names']['H'] = [
    'WID', 'VEL', 'ZH', 'ZV', 'ZDR', 'RHO', 'PHI', 'ST1', 'ST2', 'WBN', 'MPH',
    'CLT']
PMOMENTS['types'] = {'PHI':'h'} # ALl others are bytes



"""
================================================
Auxiliary input and output (:mod:`pyart.aux_io`)
================================================

.. currentmodule:: pyart.aux_io

Additional classes and functions for reading and writing data from a number
of file formats.

These auxiliary input/output routines are not as well polished as those in
:mod:`pyart.io`.  They may require addition dependencies beyond those required
for a standard Py-ART install, use non-standard function parameter and naming,
are not supported by the :py:func:`pyart.io.read` function and are not fully
tested if tested at all. Please use these at your own risk.

Bugs in these function should be reported but fixing them may not be a
priority.

Reading radar data
==================

.. autosummary::
    :toctree: generated/

    read_d3r_gcpex_nc
    read_gamic
    read_kazr
    read_noxp_iphex_nc
    read_odim_h5
    read_odim_grid_h5
    read_odim_vp_h5
    read_pattern
    read_radx
    read_rainbow_wrl
    read_metranet
    read_cartesian_metranet
    read_gif
    read_bin
    read_iq
    read_bin_mf
    read_png
    read_dat_mf
    read_grib
    read_rainbow_psr
    read_rainbow_psr_spectra
    read_spectra
    read_cf1
    read_cf1_cartesian
    read_cf1_cartesian_mf
    read_hpl

Writing radar data
==================

.. autosummary::
    :toctree: generated/

    write_odim_h5
    write_spectra

"""

from .pattern import read_pattern
from .radx import read_radx
from .d3r_gcpex_nc import read_d3r_gcpex_nc
from .noxp_iphex_nc import read_noxp_iphex_nc
from .arm_vpt import read_kazr
from .edge_netcdf import read_edge_netcdf
from .odim_h5 import read_odim_h5, read_odim_grid_h5, read_odim_vp_h5
from .odim_h5_writer import write_odim_h5
from .gamic_hdf5 import read_gamic
from .sinarame_h5 import read_sinarame_h5
from .sinarame_h5 import write_sinarame_cfradial
from .rainbow_wrl import read_rainbow_wrl
from .rainbow_psr import read_rainbow_psr, read_rainbow_psr_spectra
from .rainbow_psr import read_psr_header, read_psr_cpi_header, read_psr_spectra
from .rainbow_psr import convert_data
from .spectra import read_spectra, write_spectra
from .cf1 import read_cf1
from .cf1_cartesian import read_cf1_cartesian, read_cf1_cartesian_mf

from .metranet_reader import read_metranet
from .metranet_cartesian_reader import read_cartesian_metranet
from .rad4alp_iq_reader import read_iq

from .metranet_c import get_library
from .metranet_c import read_product as read_product_c
from .metranet_c import read_file as read_file_c
from .metranet_python import read_product as read_product_py
from .metranet_python import read_file as read_file_py

from .rad4alp_gif_reader import read_gif
from .rad4alp_bin_reader import read_bin
from .mf_bin_reader import read_bin_mf
from .mf_png_reader import read_png
from .mf_grib_reader import read_grib
from .mf_dat_reader import read_dat_mf
from .hpl_reader import read_hpl

__all__ = [s for s in dir() if not s.startswith('_')]
